#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 16 15:41:26 2020

@author: ljia

**This script demonstrates how to generate a graph preimage using Boria's method with cost matrices learning.**
"""

"""**1.   Get dataset.**"""

from gklearn.utils import Dataset, split_dataset_by_target

# Predefined dataset name, use dataset "MAO".
ds_name = 'MAO'
# The node/edge labels that will not be used in the computation.
irrelevant_labels = {'node_attrs': ['x', 'y', 'z'], 'edge_labels': ['bond_stereo']}

# Initialize a Dataset.
dataset_all = Dataset()
# Load predefined dataset "MAO".
dataset_all.load_predefined_dataset(ds_name)
# Remove irrelevant labels.
dataset_all.remove_labels(**irrelevant_labels)
# Split the whole dataset according to the classification targets.
datasets = split_dataset_by_target(dataset_all)
# Get the first class of graphs, whose median preimage will be computed.
dataset = datasets[0]
len(dataset.graphs)

"""**2.  Set parameters.**"""

import multiprocessing

# Parameters for MedianPreimageGenerator (our method).
mpg_options = {'init_method': 'random', # how to initialize node label cost vector. "random" means to initialize randomly.
			   'init_ecc': [4, 4, 2, 1, 1, 1], # initial edit costs.
			   'ds_name': ds_name, # name of the dataset.
			   'parallel': True, # @todo: whether the parallel scheme is to be used.
			   'time_limit_in_sec': 0, # maximum time limit to compute the preimage. If set to 0 then no limit.
			   'max_itrs': 3, # maximum iteration limit to optimize edit costs. If set to 0 then no limit.
			   'max_itrs_without_update': 3, # If the times that edit costs is not update is more than this number, then the optimization stops.
			   'epsilon_residual': 0.01, # In optimization, the residual is only considered changed if the change is bigger than this number.
			   'epsilon_ec': 0.1, # In optimization, the edit costs are only considered changed if the changes are bigger than this number.
			   'verbose': 2 # whether to print out results.
               }
# Parameters for graph kernel computation.
kernel_options = {'name': 'PathUpToH', # use path kernel up to length h.
				  'depth': 9,
				  'k_func': 'MinMax',
				  'compute_method': 'trie',
				  'parallel': 'imap_unordered', # or None
				  'n_jobs': multiprocessing.cpu_count(),
				  'normalize': True, # whether to use normalized Gram matrix to optimize edit costs.
				  'verbose': 2 # whether to print out results.
                  }
# Parameters for GED computation.
ged_options = {'method': 'BIPARTITE', # use Bipartite huristic.
			   'initialization_method': 'RANDOM', # or 'NODE', etc.
			   'initial_solutions': 10, # when bigger than 1, then the method is considered mIPFP.
			   'edit_cost': 'CONSTANT', # @todo: not needed. use CONSTANT cost.
			   'attr_distance': 'euclidean', # @todo: not needed. the distance between non-symbolic node/edge labels is computed by euclidean distance.
			   'ratio_runs_from_initial_solutions': 1,
			   'threads': multiprocessing.cpu_count(), # parallel threads. Do not work if mpg_options['parallel'] = False.
			   'init_option': 'LAZY_WITHOUT_SHUFFLED_COPIES' # 'EAGER_WITHOUT_SHUFFLED_COPIES'
               }
# Parameters for MedianGraphEstimator (Boria's method).
mge_options = {'init_type': 'MEDOID', # how to initial median (compute set-median). "MEDOID" is to use the graph with smallest SOD.
			   'random_inits': 10, # number of random initialization when 'init_type' = 'RANDOM'.
			   'time_limit': 600, # maximum time limit to compute the generalized median. If set to 0 then no limit.
			   'verbose': 2, # whether to print out results.
			   'refine': False # whether to refine the final SODs or not.
               }
print('done.')

"""**3.   Run median preimage generator.**"""

from gklearn.preimage import MedianPreimageGeneratorCML

# Create median preimage generator instance.
mpg = MedianPreimageGeneratorCML()
# Add dataset.
mpg.dataset = dataset
# Set parameters.
mpg.set_options(**mpg_options.copy())
mpg.kernel_options = kernel_options.copy()
mpg.ged_options = ged_options.copy()
mpg.mge_options = mge_options.copy()
# Run.
mpg.run()

"""**4. Get results.**"""

# Get results.
import pprint
pp = pprint.PrettyPrinter(indent=4) # pretty print
results = mpg.get_results()
pp.pprint(results)

# Draw generated graphs.
def draw_graph(graph):
	import matplotlib.pyplot as plt
	import networkx as nx
	plt.figure()
	pos = nx.spring_layout(graph)
	nx.draw(graph, pos, node_size=500, labels=nx.get_node_attributes(graph, 'atom_symbol'), font_color='w', width=3, with_labels=True)
	plt.show()
	plt.clf()
	plt.close()
 
draw_graph(mpg.set_median)
draw_graph(mpg.gen_median)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  1 17:02:51 2020

@author: ljia
"""

import numpy as np
from gklearn.utils import Dataset
import csv
import os
import os.path
from gklearn.preimage import RandomPreimageGenerator
from gklearn.utils import split_dataset_by_target
from gklearn.utils.graphfiles import saveGXL


def generate_random_preimages_by_class(ds_name, rpg_options, kernel_options, save_results=True, save_preimages=True, load_gm='auto', dir_save='', irrelevant_labels=None, edge_required=False, cut_range=None):	
	# 1. get dataset.
	print('1. getting dataset...')
	dataset_all = Dataset()
	dataset_all.load_predefined_dataset(ds_name)
	dataset_all.trim_dataset(edge_required=edge_required)
	if irrelevant_labels is not None:
		dataset_all.remove_labels(**irrelevant_labels)
	if cut_range is not None:
		dataset_all.cut_graphs(cut_range)
	datasets = split_dataset_by_target(dataset_all)

	if save_results:
		# create result files.
		print('creating output files...')
		fn_output_detail, fn_output_summary = _init_output_file_preimage(ds_name, kernel_options['name'], dir_save)
		

	dis_k_dataset_list = []
	dis_k_preimage_list = []
	time_precompute_gm_list = []
	time_generate_list = []
	time_total_list = []
	itrs_list = []
	num_updates_list = []
	if load_gm == 'auto':
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile_exist = os.path.isfile(os.path.abspath(gm_fname))
		if gmfile_exist:
			gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
			gram_matrix_unnorm_list = [item for item in gmfile['gram_matrix_unnorm_list']]
			time_precompute_gm_list = gmfile['run_time_list'].tolist()
		else:
			gram_matrix_unnorm_list = []
			time_precompute_gm_list = []
	elif not load_gm:
		gram_matrix_unnorm_list = []
		time_precompute_gm_list = []
	else:
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
		gram_matrix_unnorm_list = [item for item in gmfile['gram_matrix_unnorm_list']]
		time_precompute_gm_list = gmfile['run_time_list'].tolist()	
		
	print('starting generating preimage for each class of target...')
	idx_offset = 0
	for idx, dataset in enumerate(datasets):
		target = dataset.targets[0]
		print('\ntarget =', target, '\n')
#		if target != 1:
# 			continue
		
		num_graphs = len(dataset.graphs)
		if num_graphs < 2:
			print('\nnumber of graphs = ', num_graphs, ', skip.\n')
			idx_offset += 1
			continue
			
		# 2. set parameters.
		print('2. initializing mpg and setting parameters...')
		if load_gm:
			if gmfile_exist:
				rpg_options['gram_matrix_unnorm'] = gram_matrix_unnorm_list[idx - idx_offset]
				rpg_options['runtime_precompute_gm'] = time_precompute_gm_list[idx - idx_offset]
		rpg = RandomPreimageGenerator()
		rpg.dataset = dataset
		rpg.set_options(**rpg_options.copy())
		rpg.kernel_options = kernel_options.copy()

		# 3. compute preimage.
		print('3. computing preimage...')
		rpg.run()
		results = rpg.get_results()
		
		# 4. save results (and median graphs).
		print('4. saving results (and preimages)...')
		# write result detail.
		if save_results:
			print('writing results to files...')

			f_detail = open(dir_save + fn_output_detail, 'a')
			csv.writer(f_detail).writerow([ds_name, kernel_options['name'], 
					  num_graphs, target, 1, 
					  results['k_dis_dataset'], results['k_dis_preimage'],
					  results['runtime_precompute_gm'], 
					  results['runtime_generate_preimage'], results['runtime_total'],
					  results['itrs'], results['num_updates']])
			f_detail.close()
		
			# compute result summary.
			dis_k_dataset_list.append(results['k_dis_dataset'])
			dis_k_preimage_list.append(results['k_dis_preimage'])
			time_precompute_gm_list.append(results['runtime_precompute_gm'])
			time_generate_list.append(results['runtime_generate_preimage'])
			time_total_list.append(results['runtime_total'])
			itrs_list.append(results['itrs'])
			num_updates_list.append(results['num_updates'])
			
			# write result summary for each letter.
			f_summary = open(dir_save + fn_output_summary, 'a')
			csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
					  num_graphs, target,
					  results['k_dis_dataset'], results['k_dis_preimage'],
					  results['runtime_precompute_gm'], 
					  results['runtime_generate_preimage'], results['runtime_total'],
					  results['itrs'], results['num_updates']])
			f_summary.close()
			 
		# save median graphs.
		if save_preimages:
			os.makedirs(dir_save + 'preimages/', exist_ok=True)
			print('Saving preimages to files...')
			fn_best_dataset = dir_save + 'preimages/g_best_dataset.' + 'nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(rpg.best_from_dataset, fn_best_dataset + '.gxl', method='default', 
					node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
			fn_preimage = dir_save + 'preimages/g_preimage.' + 'nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(rpg.preimage, fn_preimage + '.gxl', method='default',
		  			node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
				
		if (load_gm == 'auto' and not gmfile_exist) or not load_gm:
			gram_matrix_unnorm_list.append(rpg.gram_matrix_unnorm)

	# write result summary for each class. 
	if save_results:
		dis_k_dataset_mean = np.mean(dis_k_dataset_list)
		dis_k_preimage_mean = np.mean(dis_k_preimage_list)
		time_precompute_gm_mean = np.mean(time_precompute_gm_list)
		time_generate_mean = np.mean(time_generate_list)
		time_total_mean = np.mean(time_total_list)
		itrs_mean = np.mean(itrs_list)
		num_updates_mean = np.mean(num_updates_list)
		f_summary = open(dir_save + fn_output_summary, 'a')
		csv.writer(f_summary).writerow([ds_name, kernel_options['name'],
				  num_graphs, 'all',
				  dis_k_dataset_mean, dis_k_preimage_mean,
				  time_precompute_gm_mean,
				  time_generate_mean, time_total_mean, itrs_mean, 
				  num_updates_mean])
		f_summary.close()
	
	# write Gram matrices to file.
	if (load_gm == 'auto' and not gmfile_exist) or not load_gm:
		np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm_list=gram_matrix_unnorm_list, run_time_list=time_precompute_gm_list)	

	print('\ncomplete.\n')	

	
def _init_output_file_preimage(ds_name, gkernel, dir_output):
	os.makedirs(dir_output, exist_ok=True)
	fn_output_detail = 'results_detail.' + ds_name + '.' + gkernel + '.csv'
	f_detail = open(dir_output + fn_output_detail, 'a')
	csv.writer(f_detail).writerow(['dataset', 'graph kernel', 'num graphs', 
			  'target', 'repeat', 'dis_k best from dataset', 'dis_k preimage',
			  'time precompute gm', 'time generate preimage', 'time total',
			  'itrs', 'num updates'])
	f_detail.close()
	
	fn_output_summary = 'results_summary.' + ds_name + '.' + gkernel + '.csv'
	f_summary = open(dir_output + fn_output_summary, 'a')
	csv.writer(f_summary).writerow(['dataset', 'graph kernel', 'num graphs', 
			  'target', 'dis_k best from dataset', 'dis_k preimage',
			  'time precompute gm', 'time generate preimage', 'time total',
			  'itrs', 'num updates'])
	f_summary.close()
	
	return fn_output_detail, fn_output_summary


# -*- coding: utf-8 -*-
"""compute_graph_kernel.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/17Q2QCl9CAtDweGF8LiWnWoN2laeJqT0u

**This script demonstrates how to compute a graph kernel.**
---

**0.   Install `graphkit-learn`.**
"""

"""**1.   Get dataset.**"""

from gklearn.utils import Dataset

# Predefined dataset name, use dataset "MUTAG".
ds_name = 'MUTAG'

# Initialize a Dataset.
dataset = Dataset()
# Load predefined dataset "MUTAG".
dataset.load_predefined_dataset(ds_name)
len(dataset.graphs)

"""**2.  Compute graph kernel.**"""

from gklearn.kernels import PathUpToH

# Initailize parameters for graph kernel computation.
kernel_options = {'depth': 3,
			      		  'k_func': 'MinMax',
					        'compute_method': 'trie'
								 }

# Initialize graph kernel.
graph_kernel = PathUpToH(node_labels=dataset.node_labels, # list of node label names.
						 edge_labels=dataset.edge_labels, # list of edge label names.
						 ds_infos=dataset.get_dataset_infos(keys=['directed']), # dataset information required for computation.
						 **kernel_options, # options for computation.
						 )

print('done.')

import multiprocessing
import matplotlib.pyplot as plt

# Compute Gram matrix.
gram_matrix, run_time = graph_kernel.compute(dataset.graphs,
											 parallel='imap_unordered', # or None.
											 n_jobs=multiprocessing.cpu_count(), # number of parallel jobs.
											 normalize=True, # whether to return normalized Gram matrix.
											 verbose=2 # whether to print out results.
											 )
# Print results.
print()
print(gram_matrix)
print(run_time)
plt.imshow(gram_matrix)

import multiprocessing

# Compute grah kernels between a graph and a list of graphs.
kernel_list, run_time = graph_kernel.compute(dataset.graphs, # a list of graphs. 
                                             dataset.graphs[0], # a single graph.
											 parallel='imap_unordered', # or None.
											 n_jobs=multiprocessing.cpu_count(), # number of parallel jobs.
											 verbose=2 # whether to print out results.
                                            )
# Print results.
print()
print(kernel_list)
print(run_time)

import multiprocessing

# Compute a grah kernel between two graphs.
kernel, run_time = graph_kernel.compute(dataset.graphs[0], # a single graph. 
                                        dataset.graphs[1], # another single graph.
										verbose=2 # whether to print out results.
                                       )
# Print results.
print()
print(kernel)
print(run_time)


import numpy as np

from gklearn.ged.model.distances import sum_squares, euclid_d
from gklearn.ged.model.ged_com import compute_geds


def optimize_costs_unlabeled(nb_cost_mat, dis_k_vec):
	"""
	Optimize edit costs to fit dis_k_vec according to edit operations in nb_cost_mat
	! take care that nb_cost_mat do not contains 0 lines
	:param nb_cost_mat: \in \mathbb{N}^{N x 6} encoding the number of edit operations for each pair of graph
	:param dis_k_vec: The N distances to fit
	"""
	import cvxpy as cp
	import numpy as np
	MAX_SAMPLE = 1000
	nb_cost_mat_m = np.array([[x[0], x[1], x[3], x[4]] for x in nb_cost_mat])
	dis_k_vec = np.array(dis_k_vec)
	# dis_k_vec_norm = dis_k_vec/np.max(dis_k_vec)

	# import pickle
	# pickle.dump([nb_cost_mat, dis_k_vec], open('debug', 'wb'))
	N = nb_cost_mat_m.shape[0]
	sub_sample = np.random.permutation(np.arange(N))
	sub_sample = sub_sample[:MAX_SAMPLE]

	x = cp.Variable(nb_cost_mat_m.shape[1])
	cost = cp.sum_squares(
		(nb_cost_mat_m[sub_sample, :] @ x) - dis_k_vec[sub_sample]
		)
	prob = cp.Problem(cp.Minimize(cost), [x >= 0])
	prob.solve()
	edit_costs_new = [x.value[0], x.value[1], 0, x.value[2], x.value[3], 0]
	edit_costs_new = [xi if xi > 0 else 0 for xi in edit_costs_new]
	residual = prob.value
	return edit_costs_new, residual


def optimize_costs_classif_unlabeled(nb_cost_mat, Y):
	"""
	Optimize edit costs to fit dis_k_vec according to edit operations in
	nb_cost_mat
	! take care that nb_cost_mat do not contains 0 lines
	:param nb_cost_mat: \in \mathbb{N}^{N x 6} encoding the number of edit
	operations for each pair of graph
	:param dis_k_vec: {-1,1}^N vector of common classes
	"""
	# import cvxpy as cp
	from ml import reg_log
	# import pickle
	# pickle.dump([nb_cost_mat, Y], open('debug', 'wb'))
	nb_cost_mat_m = np.array(
		[[x[0], x[1], x[3], x[4]]
		 for x in nb_cost_mat]
	)
	w, J, _ = reg_log(nb_cost_mat_m, Y, pos_contraint=True)
	edit_costs_new = [w[0], w[1], 0, w[2], w[3], 0]
	residual = J[-1]

	return edit_costs_new, residual


def optimize_costs_classif(nb_cost_mat, Y):
	"""
		Optimize edit costs to fit dis_k_vec according to edit operations in nb_cost_mat
		! take care that nb_cost_mat do not contains 0 lines
		:param nb_cost_mat: \in \mathbb{N}^{N x 6} encoding the number of edit operations for each pair of graph
		:param dis_k_vec: {-1,1}^N vector of common classes
	"""
	# import pickle
	# pickle.dump([nb_cost_mat, Y], open("test.pickle", "wb"))
	from ml import reg_log
	w, J, _ = reg_log(nb_cost_mat, Y, pos_contraint=True)
	return w, J[-1]


def optimize_costs(nb_cost_mat, dis_k_vec):
	"""
	Optimize edit costs to fit dis_k_vec according to edit operations in nb_cost_mat
	! take care that nb_cost_mat do not contains 0 lines
	:param nb_cost_mat: \in \mathbb{N}^{N x 6} encoding the number of edit operations for each pair of graph
	:param dis_k_vec: The N distances to fit

	Notes
	-----
	According to `CVXPY's official documentation
	<http://cvxr.com/cvx/doc/advanced.html#eliminating-quadratic-forms>`_:
	"One particular reformulation
	that we strongly encourage is to eliminate quadratic forms—that is, functions
	like sum_square, sum(square(.)) or quad_form—whenever it is possible to
	construct equivalent models using norm instead."

	When using the `NON-SYMBOLIC` cost method of GED and the `sum_square` function
	is used, it often happens that the
	problem is infeasible and the returned new edit costs are None. Due to this,
	the `norm` function is used instead. For more details, see this `answer
	<https://stackoverflow.com/a/65526728/9360800>`_.
	"""
	import cvxpy as cp
	x = cp.Variable(nb_cost_mat.shape[1])
	# Use `cp.norm` instead of `cp.sum_square`.
	cost = cp.norm((nb_cost_mat @ x) - dis_k_vec)
	constraints = [
		x >= [0.01 for i in range(nb_cost_mat.shape[1])],
		np.array([1.0, 1.0, -1.0, 0.0, 0.0, 0.0]).T @ x >= 0.0,
		np.array([0.0, 0.0, 0.0, 1.0, 1.0, -1.0]).T @ x >= 0.0
	]
	try:
		prob = cp.Problem(cp.Minimize(cost), constraints)
		# The tests on Redox dataset show that the solvers `cp.ECOS` and `cp.SCS`
		# yield similar results, where the default solver `cp.MOSEK` requires a
		# lisence (until 2023.03.21) (T_T). See `this answer
		# <https://stackoverflow.com/a/65526728/9360800>`_ for more details.
		prob.solve(solver=cp.ECOS)  # verbose=True)
	except cp.error.SolverError as e:
		prob = cp.Problem(cp.Minimize(cost), constraints)
		prob.solve(solver=cp.SCS)  # verbose=True)

	edit_costs_new = x.value
	residual = prob.value

	# Notice that the returned residual is the distance instead of the squared
	# distance. You may want to revised the codes where this function is invoked.
	return edit_costs_new, residual


def compute_optimal_costs(
		G, y, init_costs=[3, 3, 1, 3, 3, 1],
		y_distance=euclid_d,
		mode='reg', unlabeled=False,
		ed_method='BIPARTITE',
		edit_cost_fun='CONSTANT',
		ite_max=5,
		rescue_optim_failure=False,
		verbose=True,
		**kwargs
):
	N = len(y)

	G_pairs = []
	distances_vec = []

	for i in range(N):
		for j in range(i + 1, N):
			G_pairs.append([i, j])
			distances_vec.append(y_distance(y[i], y[j]))
	ged_vec_init, n_edit_operations = compute_geds(
		G_pairs, G, init_costs, ed_method, edit_cost_fun=edit_cost_fun,
		verbose=verbose, **kwargs
	)

	residual_list = [sum_squares(ged_vec_init, distances_vec)]

	if (mode == 'reg'):
		if unlabeled:
			method_optim = optimize_costs_unlabeled
		else:
			method_optim = optimize_costs

	elif (mode == 'classif'):
		if unlabeled:
			method_optim = optimize_costs_classif_unlabeled
		else:
			method_optim = optimize_costs_classif

	edit_costs_old = init_costs.copy()
	for i in range(ite_max):
		if verbose:
			print('ite', i + 1, '/', ite_max, ':')
		# compute GEDs and numbers of edit operations.
		edit_costs_new, residual = method_optim(
			np.array(n_edit_operations), distances_vec
		)
		# if the optimization fails:
		if edit_costs_new is None:
			if rescue_optim_failure:
				import warnings
				warnings.warn(
					'Oops (o_O)! The optimization of edit costs of this iteration '
					'(%d) seems failed and `edit_costs_new` returns `None`! '
					'Edit costs computed last iteration will be reused for the next '
					'iteration. This error may occur because the '
					'corrsponding CVXPY problem is infeasible. Please check '
					'`n_edit_operations` and `distance_vec` or ajust the constraints'
					'of the CVXPY porblem.' % (i + 1)
				)
				edit_costs_new = edit_costs_old.copy()
		else:
			edit_costs_old = edit_costs_new.copy()
		ged_vec, n_edit_operations = compute_geds(
			G_pairs, G, edit_costs_new, ed_method, edit_cost_fun=edit_cost_fun,
			verbose=verbose, **kwargs
		)
		residual_list.append(sum_squares(ged_vec, distances_vec))

	return edit_costs_new


def get_optimal_costs_GH2020(**kwargs):
	import pickle
	import os
	dir_root = 'cj/output/'
	ds_name = kwargs.get('ds_name')
	nb_trial = kwargs.get('nb_trial')
	file_name = os.path.join(
		dir_root, 'costs.' + ds_name + '.' + str(nb_trial) + '.pkl'
	)
	with open(file_name, 'rb') as f:
		edit_costs = pickle.load(f)
	return edit_costs



"""Those who are not graph kernels. We can be kernels for nodes or edges!
These kernels are defined between pairs of vectors.
"""
import numpy as np


def kronecker_delta_kernel(x, y):
	"""Delta kernel. Return 1 if x == y, 0 otherwise.

	Parameters
	----------
	x, y : any
		Two parts to compare.

	Return
	------
	kernel : integer
		Delta kernel.

	References
	----------
	[1] H. Kashima, K. Tsuda, and A. Inokuchi. Marginalized kernels between
	labeled graphs. In Proceedings of the 20th International Conference on
	Machine Learning, Washington, DC, United States, 2003.
	"""
	return (1 if np.array_equal(x, y) else 0)


def delta_kernel(x, y):
	return x == y  # (1 if condition else 0)


def deltakernel(x, y):
	return delta_kernel(x, y)


def gaussian_kernel(x, y, gamma=None):
	"""Gaussian kernel.
	Compute the rbf (gaussian) kernel between x and y:

		K(x, y) = exp(-gamma ||x-y||^2).

	Read more in the `User Guide of scikit-learn library <https://scikit-learn.org/stable/modules/metrics.html#rbf-kernel>`__.

	Parameters
	----------
	x, y : array

	gamma : float, default None
		If None, defaults to 1.0 / n_features

	Returns
	-------
	kernel : float
	"""
	if gamma is None:
		gamma = 1.0 / len(x)

	# xt = np.array([float(itm) for itm in x]) # @todo: move this to dataset or datafile to speed up.
	# yt = np.array([float(itm) for itm in y])
	# 	kernel = xt - yt
	# 	kernel = kernel ** 2
	# 	kernel = np.sum(kernel)
	# 	kernel *= -gamma
	# 	kernel = np.exp(kernel)
	# 	return kernel

	return np.exp((np.sum(np.subtract(x, y) ** 2)) * -gamma)


def tanimoto_kernel(x, y):
	xy = np.dot(x, y)
	return xy / (np.dot(x, x) + np.dot(y, y) - xy)


def gaussiankernel(x, y, gamma=None):
	return gaussian_kernel(x, y, gamma=gamma)


def polynomial_kernel(x, y, gamma=1, coef0=0, d=1):
	return (np.dot(x, y) * gamma + coef0) ** d


def highest_polynomial_kernel(x, y, d=1, c=0):
	"""Polynomial kernel.
	Compute the polynomial kernel between x and y:

		K(x, y) = <x, y> ^d + c.

	Parameters
	----------
	x, y : array

	d : integer, default 1

	c : float, default 0

	Returns
	-------
	kernel : float
	"""
	return np.dot(x, y) ** d + c


def polynomialkernel(x, y, d=1, c=0):
	return highest_polynomial_kernel(x, y, d=d, c=c)


def linear_kernel(x, y):
	"""Polynomial kernel.
	Compute the polynomial kernel between x and y:

		K(x, y) = <x, y>.

	Parameters
	----------
	x, y : array

	d : integer, default 1

	c : float, default 0

	Returns
	-------
	kernel : float
	"""
	return np.dot(x, y)


def linearkernel(x, y):
	return linear_kernel(x, y)


def cosine_kernel(x, y):
	return np.dot(x, y) / (np.linalg.norm(x) * np.linalg.norm(y))


def sigmoid_kernel(x, y, gamma=None, coef0=1):
	if gamma is None:
		gamma = 1.0 / len(x)

	k = np.dot(x, y)
	k *= gamma
	k += coef0
	k = np.tanh(k)
	# 	k = np.tanh(k, k) # compute tanh in-place
	return k


def laplacian_kernel(x, y, gamma=None):
	if gamma is None:
		gamma = 1.0 / len(x)

	k = -gamma * np.linalg.norm(np.subtract(x, y))
	k = np.exp(k)
	return k


def chi2_kernel(x, y, gamma=1.0):
	k = np.divide(np.subtract(x, y) ** 2, np.add(x, y))
	k = np.sum(k)
	k *= -gamma
	return np.exp(k)


def exponential_kernel(x, y, gamma=None):
	if gamma is None:
		gamma = 1.0 / len(x)

	return np.exp(np.dot(x, y) * gamma)


def intersection_kernel(x, y):
	return np.sum(np.minimum(x, y))


def multiquadratic_kernel(x, y, c=0):
	return np.sqrt((np.sum(np.subtract(x, y) ** 2)) + c)


def inverse_multiquadratic_kernel(x, y, c=0):
	return 1 / multiquadratic_kernel(x, y, c=c)


def kernelsum(k1, k2, d11, d12, d21=None, d22=None, lamda1=1, lamda2=1):
	"""Sum of a pair of kernels.

	k = lamda1 * k1(d11, d12) + lamda2 * k2(d21, d22)

	Parameters
	----------
	k1, k2 : function
		A pair of kernel functions.
	d11, d12:
		Inputs of k1. If d21 or d22 is None, apply d11, d12 to both k1 and k2.
	d21, d22:
		Inputs of k2.
	lamda1, lamda2: float
		Coefficients of the product.

	Return
	------
	kernel : integer

	"""
	if d21 is None or d22 is None:
		kernel = lamda1 * k1(d11, d12) + lamda2 * k2(d11, d12)
	else:
		kernel = lamda1 * k1(d11, d12) + lamda2 * k2(d21, d22)
	return kernel


def kernelproduct(k1, k2, d11, d12, d21=None, d22=None, lamda=1):
	"""Product of a pair of kernels.

	k = lamda * k1(d11, d12) * k2(d21, d22)

	Parameters
	----------
	k1, k2 : function
		A pair of kernel functions.
	d11, d12:
		Inputs of k1. If d21 or d22 is None, apply d11, d12 to both k1 and k2.
	d21, d22:
		Inputs of k2.
	lamda: float
		Coefficient of the product.

	Return
	------
	kernel : integer
	"""
	if d21 is None or d22 is None:
		kernel = lamda * k1(d11, d12) * k2(d11, d12)
	else:
		kernel = lamda * k1(d11, d12) * k2(d21, d22)
	return kernel


if __name__ == '__main__':
	o = polynomialkernel([1, 2], [3, 4], 2, 3)



""" Obtain all kinds of attributes of a graph dataset.

This file is for old version of graphkit-learn.
"""


def get_dataset_attributes(Gn,
						   target=None,
						   attr_names=[],
						   node_label=None,
						   edge_label=None):
	"""Returns the structure and property information of the graph dataset Gn.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs whose information will be returned.

	target : list
		The list of classification targets corresponding to Gn. Only works for
		classification problems.

	attr_names : list
		List of strings which indicate which informations will be returned. The
		possible choices includes:

		'substructures': sub-structures Gn contains, including 'linear', 'non
	linear' and 'cyclic'.

		'node_labeled': whether vertices have symbolic labels.

		'edge_labeled': whether egdes have symbolic labels.

		'is_directed': whether graphs in Gn are directed.

		'dataset_size': number of graphs in Gn.

		'ave_node_num': average number of vertices of graphs in Gn.

		'min_node_num': minimum number of vertices of graphs in Gn.

		'max_node_num': maximum number of vertices of graphs in Gn.

		'ave_edge_num': average number of edges of graphs in Gn.

		'min_edge_num': minimum number of edges of graphs in Gn.

		'max_edge_num': maximum number of edges of graphs in Gn.

		'ave_node_degree': average vertex degree of graphs in Gn.

		'min_node_degree': minimum vertex degree of graphs in Gn.

		'max_node_degree': maximum vertex degree of graphs in Gn.

		'ave_fill_factor': average fill factor (number_of_edges /
	(number_of_nodes ** 2)) of graphs in Gn.

		'min_fill_factor': minimum fill factor of graphs in Gn.

		'max_fill_factor': maximum fill factor of graphs in Gn.

		'node_label_num': number of symbolic vertex labels.

		'edge_label_num': number of symbolic edge labels.

		'node_attr_dim': number of dimensions of non-symbolic vertex labels.
	Extracted from the 'attributes' attribute of graph nodes.

		'edge_attr_dim': number of dimensions of non-symbolic edge labels.
	Extracted from the 'attributes' attribute of graph edges.

		'class_number': number of classes. Only available for classification problems.

	node_label : string
		Node attribute used as label. The default node label is atom. Mandatory
		when 'node_labeled' or 'node_label_num' is required.

	edge_label : string
		Edge attribute used as label. The default edge label is bond_type.
		Mandatory when 'edge_labeled' or 'edge_label_num' is required.

	Return
	------
	attrs : dict
		Value for each property.
	"""
	import networkx as nx
	import numpy as np

	attrs = {}

	def get_dataset_size(Gn):
		return len(Gn)

	def get_all_node_num(Gn):
		return [nx.number_of_nodes(G) for G in Gn]

	def get_ave_node_num(all_node_num):
		return np.mean(all_node_num)

	def get_min_node_num(all_node_num):
		return np.amin(all_node_num)

	def get_max_node_num(all_node_num):
		return np.amax(all_node_num)

	def get_all_edge_num(Gn):
		return [nx.number_of_edges(G) for G in Gn]

	def get_ave_edge_num(all_edge_num):
		return np.mean(all_edge_num)

	def get_min_edge_num(all_edge_num):
		return np.amin(all_edge_num)

	def get_max_edge_num(all_edge_num):
		return np.amax(all_edge_num)

	def is_node_labeled(Gn):
		return False if node_label is None else True

	def get_node_label_num(Gn):
		nl = set()
		for G in Gn:
			nl = nl | set(nx.get_node_attributes(G, node_label).values())
		return len(nl)

	def is_edge_labeled(Gn):
		return False if edge_label is None else True

	def get_edge_label_num(Gn):
		el = set()
		for G in Gn:
			el = el | set(nx.get_edge_attributes(G, edge_label).values())
		return len(el)

	def is_directed(Gn):
		return nx.is_directed(Gn[0])

	def get_ave_node_degree(Gn):
		return np.mean([np.mean(list(dict(G.degree()).values())) for G in Gn])

	def get_max_node_degree(Gn):
		return np.amax([np.mean(list(dict(G.degree()).values())) for G in Gn])

	def get_min_node_degree(Gn):
		return np.amin([np.mean(list(dict(G.degree()).values())) for G in Gn])

	# get fill factor, the number of non-zero entries in the adjacency matrix.
	def get_ave_fill_factor(Gn):
		return np.mean([nx.number_of_edges(G) / (nx.number_of_nodes(G)
										   * nx.number_of_nodes(G)) for G in Gn])

	def get_max_fill_factor(Gn):
		return np.amax([nx.number_of_edges(G) / (nx.number_of_nodes(G)
										   * nx.number_of_nodes(G)) for G in Gn])

	def get_min_fill_factor(Gn):
		return np.amin([nx.number_of_edges(G) / (nx.number_of_nodes(G)
										   * nx.number_of_nodes(G)) for G in Gn])

	def get_substructures(Gn):
		subs = set()
		for G in Gn:
			degrees = list(dict(G.degree()).values())
			if any(i == 2 for i in degrees):
				subs.add('linear')
			if np.amax(degrees) >= 3:
				subs.add('non linear')
			if 'linear' in subs and 'non linear' in subs:
				break

		if is_directed(Gn):
			for G in Gn:
				if len(list(nx.find_cycle(G))) > 0:
					subs.add('cyclic')
					break
# 		else:
# 			# @todo: this method does not work for big graph with large amount of edges like D&D, try a better way.
# 			upper = np.amin([nx.number_of_edges(G) for G in Gn]) * 2 + 10
# 			for G in Gn:
# 				if (nx.number_of_edges(G) < upper):
# 					cyc = list(nx.simple_cycles(G.to_directed()))
# 					if any(len(i) > 2 for i in cyc):
# 						subs.add('cyclic')
# 						break
# 			if 'cyclic' not in subs:
# 				for G in Gn:
# 					cyc = list(nx.simple_cycles(G.to_directed()))
# 					if any(len(i) > 2 for i in cyc):
# 						subs.add('cyclic')
# 						break

		return subs

	def get_class_num(target):
		return len(set(target))

	def get_node_attr_dim(Gn):
		for G in Gn:
			for n in G.nodes(data=True):
				if 'attributes' in n[1]:
					return len(n[1]['attributes'])
		return 0

	def get_edge_attr_dim(Gn):
		for G in Gn:
			if nx.number_of_edges(G) > 0:
				for e in G.edges(data=True):
					if 'attributes' in e[2]:
						return len(e[2]['attributes'])
		return 0

	if attr_names == []:
		attr_names = [
			'substructures',
			'node_labeled',
			'edge_labeled',
			'is_directed',
			'dataset_size',
			'ave_node_num',
			'min_node_num',
			'max_node_num',
			'ave_edge_num',
			'min_edge_num',
			'max_edge_num',
			'ave_node_degree',
			'min_node_degree',
			'max_node_degree',
			'ave_fill_factor',
			'min_fill_factor',
			'max_fill_factor',
			'node_label_num',
			'edge_label_num',
			'node_attr_dim',
			'edge_attr_dim',
			'class_number',
		]

	# dataset size
	if 'dataset_size' in attr_names:

		attrs.update({'dataset_size': get_dataset_size(Gn)})

	# graph node number
	if any(i in attr_names
		   for i in ['ave_node_num', 'min_node_num', 'max_node_num']):

		all_node_num = get_all_node_num(Gn)

	if 'ave_node_num' in attr_names:

		attrs.update({'ave_node_num': get_ave_node_num(all_node_num)})

	if 'min_node_num' in attr_names:

		attrs.update({'min_node_num': get_min_node_num(all_node_num)})

	if 'max_node_num' in attr_names:

		attrs.update({'max_node_num': get_max_node_num(all_node_num)})

	# graph edge number
	if any(i in attr_names for i in
		   ['ave_edge_num', 'min_edge_num', 'max_edge_num']):

		all_edge_num = get_all_edge_num(Gn)

	if 'ave_edge_num' in attr_names:

		attrs.update({'ave_edge_num': get_ave_edge_num(all_edge_num)})

	if 'max_edge_num' in attr_names:

		attrs.update({'max_edge_num': get_max_edge_num(all_edge_num)})

	if 'min_edge_num' in attr_names:

		attrs.update({'min_edge_num': get_min_edge_num(all_edge_num)})

	# label number
	if any(i in attr_names for i in ['node_labeled', 'node_label_num']):
		is_nl = is_node_labeled(Gn)
		node_label_num = get_node_label_num(Gn)

	if 'node_labeled' in attr_names:
		# graphs are considered node unlabeled if all nodes have the same label.
		attrs.update({'node_labeled': is_nl if node_label_num > 1 else False})

	if 'node_label_num' in attr_names:
		attrs.update({'node_label_num': node_label_num})

	if any(i in attr_names for i in ['edge_labeled', 'edge_label_num']):
		is_el = is_edge_labeled(Gn)
		edge_label_num = get_edge_label_num(Gn)

	if 'edge_labeled' in attr_names:
		# graphs are considered edge unlabeled if all edges have the same label.
		attrs.update({'edge_labeled': is_el if edge_label_num > 1 else False})

	if 'edge_label_num' in attr_names:
		attrs.update({'edge_label_num': edge_label_num})

	if 'is_directed' in attr_names:
		attrs.update({'is_directed': is_directed(Gn)})

	if 'ave_node_degree' in attr_names:
		attrs.update({'ave_node_degree': get_ave_node_degree(Gn)})

	if 'max_node_degree' in attr_names:
		attrs.update({'max_node_degree': get_max_node_degree(Gn)})

	if 'min_node_degree' in attr_names:
		attrs.update({'min_node_degree': get_min_node_degree(Gn)})

	if 'ave_fill_factor' in attr_names:
		attrs.update({'ave_fill_factor': get_ave_fill_factor(Gn)})

	if 'max_fill_factor' in attr_names:
		attrs.update({'max_fill_factor': get_max_fill_factor(Gn)})

	if 'min_fill_factor' in attr_names:
		attrs.update({'min_fill_factor': get_min_fill_factor(Gn)})

	if 'substructures' in attr_names:
		attrs.update({'substructures': get_substructures(Gn)})

	if 'class_number' in attr_names:
		attrs.update({'class_number': get_class_num(target)})

	if 'node_attr_dim' in attr_names:
		attrs['node_attr_dim'] = get_node_attr_dim(Gn)

	if 'edge_attr_dim' in attr_names:
		attrs['edge_attr_dim'] = get_edge_attr_dim(Gn)

	from collections import OrderedDict
	return OrderedDict(
		sorted(attrs.items(), key=lambda i: attr_names.index(i[0])))


def load_predefined_dataset(ds_name):
	import os
	from gklearn.utils.graphfiles import loadDataset

	current_path = os.path.dirname(os.path.realpath(__file__)) + '/'
	if ds_name == 'Acyclic':
		ds_file = current_path + '../../datasets/Acyclic/dataset_bps.ds'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'AIDS':
		ds_file = current_path + '../../datasets/AIDS/AIDS_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Alkane':
		ds_file = current_path + '../../datasets/Alkane/dataset.ds'
		fn_targets = current_path + '../../datasets/Alkane/dataset_boiling_point_names.txt'
		graphs, targets = loadDataset(ds_file, filename_y=fn_targets)
	elif ds_name == 'COIL-DEL':
		ds_file = current_path + '../../datasets/COIL-DEL/COIL-DEL_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'COIL-RAG':
		ds_file = current_path + '../../datasets/COIL-RAG/COIL-RAG_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'COLORS-3':
		ds_file = current_path + '../../datasets/COLORS-3/COLORS-3_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Cuneiform':
		ds_file = current_path + '../../datasets/Cuneiform/Cuneiform_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'DD':
		ds_file = current_path + '../../datasets/DD/DD_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'ENZYMES':
		ds_file = current_path + '../../datasets/ENZYMES_txt/ENZYMES_A_sparse.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Fingerprint':
		ds_file = current_path + '../../datasets/Fingerprint/Fingerprint_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'FRANKENSTEIN':
		ds_file = current_path + '../../datasets/FRANKENSTEIN/FRANKENSTEIN_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Letter-high': # node non-symb
		ds_file = current_path + '../../datasets/Letter-high/Letter-high_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Letter-low': # node non-symb
		ds_file = current_path + '../../datasets/Letter-low/Letter-low_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Letter-med': # node non-symb
		ds_file = current_path + '../../datasets/Letter-med/Letter-med_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'MAO':
		ds_file = current_path + '../../datasets/MAO/dataset.ds'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Monoterpenoides':
		ds_file = current_path + '../../datasets/Monoterpenoides/dataset_10+.ds'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'MUTAG':
		ds_file = current_path + '../../datasets/MUTAG/MUTAG_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'NCI1':
		ds_file = current_path + '../../datasets/NCI1/NCI1_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'NCI109':
		ds_file = current_path + '../../datasets/NCI109/NCI109_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'PAH':
		ds_file = current_path + '../../datasets/PAH/dataset.ds'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'SYNTHETIC':
		pass
	elif ds_name == 'SYNTHETICnew':
		ds_file = current_path + '../../datasets/SYNTHETICnew/SYNTHETICnew_A.txt'
		graphs, targets = loadDataset(ds_file)
	elif ds_name == 'Synthie':
		pass
	else:
		raise Exception('The dataset name "', ds_name, '" is not pre-defined.')

	return graphs, targets


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 15 18:22:34 2020

@author: ljia
"""

import os
EXP_ROOT = os.path.dirname(os.path.realpath(__file__)) + '/'
DATASET_ROOT = os.path.dirname(os.path.realpath(__file__)) + '/datasets/'


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 11:42:48 2020

@author: ljia
"""
import numpy as np
import cvxpy as cp
import time
from gklearn.ged.learning.costs_learner import CostsLearner
from gklearn.ged.util import compute_geds_cml


class CostMatricesLearner(CostsLearner):
	
	
	def __init__(self, edit_cost='CONSTANT', triangle_rule=False, allow_zeros=True, parallel=False, verbose=2):
		super().__init__(parallel, verbose)
		self._edit_cost = edit_cost
		self._triangle_rule = triangle_rule
		self._allow_zeros = allow_zeros
	
	
	def fit(self, X, y):
		if self._edit_cost == 'LETTER':
			raise Exception('Cannot compute for cost "LETTER".')
		elif self._edit_cost == 'LETTER2':
			raise Exception('Cannot compute for cost "LETTER2".')
		elif self._edit_cost == 'NON_SYMBOLIC':
			raise Exception('Cannot compute for cost "NON_SYMBOLIC".')
		elif self._edit_cost == 'CONSTANT': # @todo: node/edge may not labeled.
			if not self._triangle_rule and self._allow_zeros:
				w = cp.Variable(X.shape[1])
				cost_fun = cp.sum_squares(X @ w - y)
				constraints = [w >= [0.0 for i in range(X.shape[1])]]
				prob = cp.Problem(cp.Minimize(cost_fun), constraints)
				self.execute_cvx(prob)
				edit_costs_new = w.value
				residual = np.sqrt(prob.value)
			elif self._triangle_rule and self._allow_zeros: # @todo
				x = cp.Variable(nb_cost_mat.shape[1])
				cost_fun = cp.sum_squares(nb_cost_mat @ x - dis_k_vec)
				constraints = [x >= [0.0 for i in range(nb_cost_mat.shape[1])],
							   np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0]).T@x >= 0.01,
							   np.array([0.0, 1.0, 0.0, 0.0, 0.0, 0.0]).T@x >= 0.01,
							   np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]).T@x >= 0.01,
							   np.array([0.0, 0.0, 0.0, 0.0, 1.0, 0.0]).T@x >= 0.01,
							   np.array([1.0, 1.0, -1.0, 0.0, 0.0, 0.0]).T@x >= 0.0,
							   np.array([0.0, 0.0, 0.0, 1.0, 1.0, -1.0]).T@x >= 0.0]
				prob = cp.Problem(cp.Minimize(cost_fun), constraints)
				self._execute_cvx(prob)
				edit_costs_new = x.value
				residual = np.sqrt(prob.value)
			elif not self._triangle_rule and not self._allow_zeros: # @todo
				x = cp.Variable(nb_cost_mat.shape[1])
				cost_fun = cp.sum_squares(nb_cost_mat @ x - dis_k_vec)
				constraints = [x >= [0.01 for i in range(nb_cost_mat.shape[1])]]
				prob = cp.Problem(cp.Minimize(cost_fun), constraints)
				self._execute_cvx(prob)
				edit_costs_new = x.value
				residual = np.sqrt(prob.value)
			elif self._triangle_rule and not self._allow_zeros: # @todo
				x = cp.Variable(nb_cost_mat.shape[1])
				cost_fun = cp.sum_squares(nb_cost_mat @ x - dis_k_vec)
				constraints = [x >= [0.01 for i in range(nb_cost_mat.shape[1])],
							   np.array([1.0, 1.0, -1.0, 0.0, 0.0, 0.0]).T@x >= 0.0,
							   np.array([0.0, 0.0, 0.0, 1.0, 1.0, -1.0]).T@x >= 0.0]
				prob = cp.Problem(cp.Minimize(cost_fun), constraints)
				self._execute_cvx(prob)
				edit_costs_new = x.value
				residual = np.sqrt(prob.value)
		else:
			raise Exception('The edit cost "', self._ged_options['edit_cost'], '" is not supported for update progress.')
		
		self._cost_list.append(edit_costs_new)
		
		
	def init_geds_and_nb_eo(self, y, graphs):
		time0 = time.time()
		self._cost_list.append(np.concatenate((self._ged_options['node_label_costs'], 
										 self._ged_options['edge_label_costs'])))
		ged_vec, self._nb_eo = self.compute_geds_and_nb_eo(graphs)
		self._residual_list.append(np.sqrt(np.sum(np.square(np.array(ged_vec) - y))))
		self._runtime_list.append(time.time() - time0)

		if self._verbose >= 2:
			print('Current node label costs:', self._cost_list[-1][0:len(self._ged_options['node_label_costs'])])
			print('Current edge label costs:', self._cost_list[-1][len(self._ged_options['node_label_costs']):])
			print('Residual list:', self._residual_list) 
	
	
	def update_geds_and_nb_eo(self, y, graphs, time0):
		self._ged_options['node_label_costs'] = self._cost_list[-1][0:len(self._ged_options['node_label_costs'])]
		self._ged_options['edge_label_costs'] = self._cost_list[-1][len(self._ged_options['node_label_costs']):]
		ged_vec, self._nb_eo = self.compute_geds_and_nb_eo(graphs)
		self._residual_list.append(np.sqrt(np.sum(np.square(np.array(ged_vec) - y))))
		self._runtime_list.append(time.time() - time0)
		
	
	def compute_geds_and_nb_eo(self, graphs):
		ged_vec, ged_mat, n_edit_operations = compute_geds_cml(graphs, options=self._ged_options, parallel=self._parallel, verbose=(self._verbose > 1))
		return ged_vec, np.array(n_edit_operations)
	
	
	def check_convergency(self):
		self._ec_changed = False
		for i, cost in enumerate(self._cost_list[-1]):
			if cost == 0:
				if self._cost_list[-2][i] > self._epsilon_ec:
					self._ec_changed = True
					break
			elif abs(cost - self._cost_list[-2][i]) / cost > self._epsilon_ec:
				self._ec_changed = True
				break
# 				if abs(cost - edit_cost_list[-2][i]) > self._epsilon_ec:
#  					ec_changed = True
#  					break
		self._residual_changed = False
		if self._residual_list[-1] == 0:
			if self._residual_list[-2] > self._epsilon_residual:
				self._residual_changed = True
		elif abs(self._residual_list[-1] - self._residual_list[-2]) / self._residual_list[-1] > self._epsilon_residual:
			self._residual_changed = True
		self._converged = not (self._ec_changed or self._residual_changed)
		if self._converged:
			self._itrs_without_update += 1
		else:
			self._itrs_without_update = 0
			self._num_updates_ecs += 1
	
	
	def print_current_states(self):
		print()
		print('-------------------------------------------------------------------------')
		print('States of iteration', self._itrs + 1)
		print('-------------------------------------------------------------------------')
# 				print('Time spend:', self._runtime_optimize_ec)
		print('Total number of iterations for optimizing:', self._itrs + 1)
		print('Total number of updating edit costs:', self._num_updates_ecs)
		print('Was optimization of edit costs converged:', self._converged)
		print('Did edit costs change:', self._ec_changed)
		print('Did residual change:', self._residual_changed)
		print('Iterations without update:', self._itrs_without_update)
		print('Current node label costs:', self._cost_list[-1][0:len(self._ged_options['node_label_costs'])])
		print('Current edge label costs:', self._cost_list[-1][len(self._ged_options['node_label_costs']):])
		print('Residual list:', self._residual_list)
		print('-------------------------------------------------------------------------')


"""
@author: linlin

@references:

	[1] Shervashidze N, Schweitzer P, Leeuwen EJ, Mehlhorn K, Borgwardt KM. 
	Weisfeiler-lehman graph kernels. Journal of Machine Learning Research. 
	2011;12(Sep):2539-61.
"""

import sys
from collections import Counter
from functools import partial
import time
#from multiprocessing import Pool
from tqdm import tqdm

import networkx as nx
import numpy as np

#from gklearn.kernels.pathKernel import pathkernel
from gklearn.utils.graphdataset import get_dataset_attributes
from gklearn.utils.parallel import parallel_gm

# @todo: support edge kernel, sp kernel, user-defined kernel.
def weisfeilerlehmankernel(*args, 
						   node_label='atom',
						   edge_label='bond_type',
						   height=0,
						   base_kernel='subtree',
						   parallel=None,
						   n_jobs=None, 
						   chunksize=None,
						   verbose=True):
	"""Compute Weisfeiler-Lehman kernels between graphs.
	
	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.
	
	G1, G2 : NetworkX graphs
		Two graphs between which the kernel is computed.		

	node_label : string
		Node attribute used as label. The default node label is atom.		

	edge_label : string
		Edge attribute used as label. The default edge label is bond_type.		

	height : int
		Subtree height.

	base_kernel : string
		Base kernel used in each iteration of WL kernel. Only default 'subtree' 
		kernel can be applied for now.

	parallel : None
		Which paralleliztion method is applied to compute the kernel. No 
		parallelization can be applied for now.

	n_jobs : int
		Number of jobs for parallelization. The default is to use all 
		computational cores. This argument is only valid when one of the 
		parallelization method is applied and can be ignored for now.

	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.

	Notes
	-----
	This function now supports WL subtree kernel only.
	"""
#		The default base 
#		kernel is subtree kernel. For user-defined kernel, base_kernel is the 
#		name of the base kernel function used in each iteration of WL kernel. 
#		This function returns a Numpy matrix, each element of which is the 
#		user-defined Weisfeiler-Lehman kernel between 2 praphs.
	# pre-process
	base_kernel = base_kernel.lower()
	Gn = args[0] if len(args) == 1 else [args[0], args[1]] # arrange all graphs in a list
	Gn = [g.copy() for g in Gn]
	ds_attrs = get_dataset_attributes(Gn, attr_names=['node_labeled'], 
									  node_label=node_label)
	if not ds_attrs['node_labeled']:
		for G in Gn:
			nx.set_node_attributes(G, '0', 'atom')

	start_time = time.time()

	# for WL subtree kernel
	if base_kernel == 'subtree':		   
		Kmatrix = _wl_kernel_do(Gn, node_label, edge_label, height, parallel, n_jobs, chunksize, verbose)

	# for WL shortest path kernel
	elif base_kernel == 'sp':
		Kmatrix = _wl_spkernel_do(Gn, node_label, edge_label, height)

	# for WL edge kernel
	elif base_kernel == 'edge':
		Kmatrix = _wl_edgekernel_do(Gn, node_label, edge_label, height)

	# for user defined base kernel
	else:
		Kmatrix = _wl_userkernel_do(Gn, node_label, edge_label, height, base_kernel)

	run_time = time.time() - start_time
	if verbose:
		print("\n --- Weisfeiler-Lehman %s kernel matrix of size %d built in %s seconds ---" 
			  % (base_kernel, len(args[0]), run_time))

	return Kmatrix, run_time


def _wl_kernel_do(Gn, node_label, edge_label, height, parallel, n_jobs, chunksize, verbose):
	"""Compute Weisfeiler-Lehman kernels between graphs.

	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.	   
	node_label : string
		node attribute used as label.
	edge_label : string
		edge attribute used as label.	  
	height : int
		wl height.

	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
	"""
	height = int(height)
	Kmatrix = np.zeros((len(Gn), len(Gn)))

	# initial for height = 0
	all_num_of_each_label = [] # number of occurence of each label in each graph in this iteration

	# for each graph
	for G in Gn:
		# get the set of original labels
		labels_ori = list(nx.get_node_attributes(G, node_label).values())
		# number of occurence of each label in G
		all_num_of_each_label.append(dict(Counter(labels_ori)))

	# Compute subtree kernel with the 0th iteration and add it to the final kernel
	compute_kernel_matrix(Kmatrix, all_num_of_each_label, Gn, parallel, n_jobs, chunksize, False)

	# iterate each height
	for h in range(1, height + 1):
		all_set_compressed = {} # a dictionary mapping original labels to new ones in all graphs in this iteration
		num_of_labels_occured = 0 # number of the set of letters that occur before as node labels at least once in all graphs
#		all_labels_ori = set() # all unique orignal labels in all graphs in this iteration
		all_num_of_each_label = [] # number of occurence of each label in G

#		# for each graph
#		# ---- use pool.imap_unordered to parallel and track progress. ----
#		pool = Pool(n_jobs)
#		itr = zip(Gn, range(0, len(Gn)))
#		if len(Gn) < 100 * n_jobs:
#			chunksize = int(len(Gn) / n_jobs) + 1
#		else:
#			chunksize = 100
#		all_multisets_list = [[] for _ in range(len(Gn))]
##		set_unique_list = [[] for _ in range(len(Gn))]
#		get_partial = partial(wrapper_wl_iteration, node_label)
##		if verbose:
##			iterator = tqdm(pool.imap_unordered(get_partial, itr, chunksize),
##							desc='wl iteration', file=sys.stdout)
##		else:
#		iterator = pool.imap_unordered(get_partial, itr, chunksize)
#		for i, all_multisets in iterator:
#			all_multisets_list[i] = all_multisets
##			set_unique_list[i] = set_unique
##			all_set_unique = all_set_unique | set(set_unique)
#		pool.close()
#		pool.join()
		
#		all_set_unique = set()
#		for uset in all_multisets_list:
#			all_set_unique = all_set_unique | set(uset)
#			
#		all_set_unique = list(all_set_unique)
##		# a dictionary mapping original labels to new ones. 
##		set_compressed = {}
##		for idx, uset in enumerate(all_set_unique):
##			set_compressed.update({uset: idx})
#			
#		for ig, G in enumerate(Gn):
#
##			# a dictionary mapping original labels to new ones. 
##			set_compressed = {}
##			# if a label occured before, assign its former compressed label, 
##			# else assign the number of labels occured + 1 as the compressed label. 
##			for value in set_unique_list[i]:
##				if uset in all_set_unique:
##					set_compressed.update({uset: all_set_compressed[value]})
##				else:
##					set_compressed.update({value: str(num_of_labels_occured + 1)})
##					num_of_labels_occured += 1
#					
##			all_set_compressed.update(set_compressed)
#			
#			# relabel nodes
#			for idx, node in enumerate(G.nodes()):
#				G.nodes[node][node_label] = all_set_unique.index(all_multisets_list[ig][idx])
#				
#			# get the set of compressed labels
#			labels_comp = list(nx.get_node_attributes(G, node_label).values())
##			all_labels_ori.update(labels_comp)
#			all_num_of_each_label[ig] = dict(Counter(labels_comp))
			
			

		
#		all_set_unique = list(all_set_unique)
		
		
		# @todo: parallel this part.
		for idx, G in enumerate(Gn):

			all_multisets = []
			for node, attrs in G.nodes(data=True):
				# Multiset-label determination.
				multiset = [G.nodes[neighbors][node_label] for neighbors in G[node]]
				# sorting each multiset
				multiset.sort()
				multiset = [attrs[node_label]] + multiset # add the prefix 
				all_multisets.append(tuple(multiset))

			# label compression
			set_unique = list(set(all_multisets)) # set of unique multiset labels
			# a dictionary mapping original labels to new ones. 
			set_compressed = {}
			# if a label occured before, assign its former compressed label, 
			# else assign the number of labels occured + 1 as the compressed label. 
			for value in set_unique:
				if value in all_set_compressed.keys():
					set_compressed.update({value: all_set_compressed[value]})
				else:
					set_compressed.update({value: str(num_of_labels_occured + 1)})
					num_of_labels_occured += 1

			all_set_compressed.update(set_compressed)

			# relabel nodes
			for idx, node in enumerate(G.nodes()):
				G.nodes[node][node_label] = set_compressed[all_multisets[idx]]

			# get the set of compressed labels
			labels_comp = list(nx.get_node_attributes(G, node_label).values())
#			all_labels_ori.update(labels_comp)
			all_num_of_each_label.append(dict(Counter(labels_comp)))

		# Compute subtree kernel with h iterations and add it to the final kernel
		compute_kernel_matrix(Kmatrix, all_num_of_each_label, Gn, parallel, n_jobs, chunksize, False)

	return Kmatrix


def wl_iteration(G, node_label):
	all_multisets = []
	for node, attrs in G.nodes(data=True):
		# Multiset-label determination.
		multiset = [G.nodes[neighbors][node_label] for neighbors in G[node]]
		# sorting each multiset
		multiset.sort()
		multiset = [attrs[node_label]] + multiset # add the prefix 
		all_multisets.append(tuple(multiset))
#	# label compression
#	set_unique = list(set(all_multisets)) # set of unique multiset labels
	return all_multisets
	
#	# a dictionary mapping original labels to new ones. 
#	set_compressed = {}
#	# if a label occured before, assign its former compressed label, 
#	# else assign the number of labels occured + 1 as the compressed label. 
#	for value in set_unique:
#		if value in all_set_compressed.keys():
#			set_compressed.update({value: all_set_compressed[value]})
#		else:
#			set_compressed.update({value: str(num_of_labels_occured + 1)})
#			num_of_labels_occured += 1
#
#	all_set_compressed.update(set_compressed)
#
#	# relabel nodes
#	for idx, node in enumerate(G.nodes()):
#		G.nodes[node][node_label] = set_compressed[all_multisets[idx]]
#
#	# get the set of compressed labels
#	labels_comp = list(nx.get_node_attributes(G, node_label).values())
#	all_labels_ori.update(labels_comp)
#	all_num_of_each_label.append(dict(Counter(labels_comp)))
#	return


def wrapper_wl_iteration(node_label, itr_item):
	g = itr_item[0]
	i = itr_item[1]
	all_multisets = wl_iteration(g, node_label)
	return i, all_multisets


def compute_kernel_matrix(Kmatrix, all_num_of_each_label, Gn, parallel, n_jobs, chunksize, verbose):
	"""Compute kernel matrix using the base kernel.
	"""
	if parallel == 'imap_unordered':
		# compute kernels.
		def init_worker(alllabels_toshare):
			global G_alllabels
			G_alllabels = alllabels_toshare
		do_partial = partial(wrapper_compute_subtree_kernel, Kmatrix)
		parallel_gm(do_partial, Kmatrix, Gn, init_worker=init_worker, 
					glbv=(all_num_of_each_label,), n_jobs=n_jobs, chunksize=chunksize, verbose=verbose)
	elif parallel is None:
		for i in range(len(Kmatrix)):
			for j in range(i, len(Kmatrix)):
				Kmatrix[i][j] = compute_subtree_kernel(all_num_of_each_label[i],
					   all_num_of_each_label[j], Kmatrix[i][j])
				Kmatrix[j][i] = Kmatrix[i][j]


def compute_subtree_kernel(num_of_each_label1, num_of_each_label2, kernel):
	"""Compute the subtree kernel.
	"""
	labels = set(list(num_of_each_label1.keys()) + list(num_of_each_label2.keys()))
	vector1 = np.array([(num_of_each_label1[label] 
						if (label in num_of_each_label1.keys()) else 0) 
						for label in labels])
	vector2 = np.array([(num_of_each_label2[label] 
						if (label in num_of_each_label2.keys()) else 0) 
						for label in labels])
	kernel += np.dot(vector1, vector2)
	return kernel


def wrapper_compute_subtree_kernel(Kmatrix, itr):
	i = itr[0]
	j = itr[1]
	return i, j, compute_subtree_kernel(G_alllabels[i], G_alllabels[j], Kmatrix[i][j])
				

def _wl_spkernel_do(Gn, node_label, edge_label, height):
	"""Compute Weisfeiler-Lehman shortest path kernels between graphs.
	
	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.	   
	node_label : string
		node attribute used as label.	  
	edge_label : string
		edge attribute used as label.	   
	height : int
		subtree height.
		
	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
	"""
	pass
	from gklearn.utils.utils import getSPGraph
	  
	# init.
	height = int(height)
	Kmatrix = np.zeros((len(Gn), len(Gn))) # init kernel

	Gn = [ getSPGraph(G, edge_weight = edge_label) for G in Gn ] # get shortest path graphs of Gn
	
	# initial for height = 0
	for i in range(0, len(Gn)):
		for j in range(i, len(Gn)):
			for e1 in Gn[i].edges(data = True):
				for e2 in Gn[j].edges(data = True):		  
					if e1[2]['cost'] != 0 and e1[2]['cost'] == e2[2]['cost'] and ((e1[0] == e2[0] and e1[1] == e2[1]) or (e1[0] == e2[1] and e1[1] == e2[0])):
						Kmatrix[i][j] += 1
			Kmatrix[j][i] = Kmatrix[i][j]
			
	# iterate each height
	for h in range(1, height + 1):
		all_set_compressed = {} # a dictionary mapping original labels to new ones in all graphs in this iteration
		num_of_labels_occured = 0 # number of the set of letters that occur before as node labels at least once in all graphs
		for G in Gn: # for each graph
			set_multisets = []
			for node in G.nodes(data = True):
				# Multiset-label determination.
				multiset = [ G.node[neighbors][node_label] for neighbors in G[node[0]] ]
				# sorting each multiset
				multiset.sort()
				multiset = node[1][node_label] + ''.join(multiset) # concatenate to a string and add the prefix 
				set_multisets.append(multiset)		  

			# label compression
			set_unique = list(set(set_multisets)) # set of unique multiset labels
			# a dictionary mapping original labels to new ones. 
			set_compressed = {}
			# if a label occured before, assign its former compressed label, else assign the number of labels occured + 1 as the compressed label 
			for value in set_unique:
				if value in all_set_compressed.keys():
					set_compressed.update({ value : all_set_compressed[value] })
				else:
					set_compressed.update({ value : str(num_of_labels_occured + 1) })
					num_of_labels_occured += 1

			all_set_compressed.update(set_compressed)
			
			# relabel nodes
			for node in G.nodes(data = True):
				node[1][node_label] = set_compressed[set_multisets[node[0]]]
				
		# Compute subtree kernel with h iterations and add it to the final kernel
		for i in range(0, len(Gn)):
			for j in range(i, len(Gn)):
				for e1 in Gn[i].edges(data = True):
					for e2 in Gn[j].edges(data = True):		  
						if e1[2]['cost'] != 0 and e1[2]['cost'] == e2[2]['cost'] and ((e1[0] == e2[0] and e1[1] == e2[1]) or (e1[0] == e2[1] and e1[1] == e2[0])):
							Kmatrix[i][j] += 1
				Kmatrix[j][i] = Kmatrix[i][j]
		
	return Kmatrix



def _wl_edgekernel_do(Gn, node_label, edge_label, height):
	"""Compute Weisfeiler-Lehman edge kernels between graphs.
	
	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.	   
	node_label : string
		node attribute used as label.	  
	edge_label : string
		edge attribute used as label.	   
	height : int
		subtree height.
		
	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
	"""	  
	pass
	# init.
	height = int(height)
	Kmatrix = np.zeros((len(Gn), len(Gn))) # init kernel
  
	# initial for height = 0
	for i in range(0, len(Gn)):
		for j in range(i, len(Gn)):
			for e1 in Gn[i].edges(data = True):
				for e2 in Gn[j].edges(data = True):		  
					if e1[2][edge_label] == e2[2][edge_label] and ((e1[0] == e2[0] and e1[1] == e2[1]) or (e1[0] == e2[1] and e1[1] == e2[0])):
						Kmatrix[i][j] += 1
			Kmatrix[j][i] = Kmatrix[i][j]
			
	# iterate each height
	for h in range(1, height + 1):
		all_set_compressed = {} # a dictionary mapping original labels to new ones in all graphs in this iteration
		num_of_labels_occured = 0 # number of the set of letters that occur before as node labels at least once in all graphs
		for G in Gn: # for each graph
			set_multisets = []			
			for node in G.nodes(data = True):
				# Multiset-label determination.
				multiset = [ G.node[neighbors][node_label] for neighbors in G[node[0]] ]
				# sorting each multiset
				multiset.sort()
				multiset = node[1][node_label] + ''.join(multiset) # concatenate to a string and add the prefix 
				set_multisets.append(multiset)		  

			# label compression
			set_unique = list(set(set_multisets)) # set of unique multiset labels
			# a dictionary mapping original labels to new ones. 
			set_compressed = {}
			# if a label occured before, assign its former compressed label, else assign the number of labels occured + 1 as the compressed label 
			for value in set_unique:
				if value in all_set_compressed.keys():
					set_compressed.update({ value : all_set_compressed[value] })
				else:
					set_compressed.update({ value : str(num_of_labels_occured + 1) })
					num_of_labels_occured += 1

			all_set_compressed.update(set_compressed)
			
			# relabel nodes
			for node in G.nodes(data = True):
				node[1][node_label] = set_compressed[set_multisets[node[0]]]
				
		# Compute subtree kernel with h iterations and add it to the final kernel
		for i in range(0, len(Gn)):
			for j in range(i, len(Gn)):
				for e1 in Gn[i].edges(data = True):
					for e2 in Gn[j].edges(data = True):		  
						if e1[2][edge_label] == e2[2][edge_label] and ((e1[0] == e2[0] and e1[1] == e2[1]) or (e1[0] == e2[1] and e1[1] == e2[0])):
							Kmatrix[i][j] += 1
				Kmatrix[j][i] = Kmatrix[i][j]
		
	return Kmatrix


def _wl_userkernel_do(Gn, node_label, edge_label, height, base_kernel):
	"""Compute Weisfeiler-Lehman kernels based on user-defined kernel between graphs.
	
	Parameters
	----------
	Gn : List of NetworkX graph
		List of graphs between which the kernels are computed.	   
	node_label : string
		node attribute used as label.	  
	edge_label : string
		edge attribute used as label.	   
	height : int
		subtree height.
	base_kernel : string
		Name of the base kernel function used in each iteration of WL kernel. This function returns a Numpy matrix, each element of which is the user-defined Weisfeiler-Lehman kernel between 2 praphs.
		
	Return
	------
	Kmatrix : Numpy matrix
		Kernel matrix, each element of which is the Weisfeiler-Lehman kernel between 2 praphs.
	"""	  
	pass
	# init.
	height = int(height)
	Kmatrix = np.zeros((len(Gn), len(Gn))) # init kernel
  
	# initial for height = 0
	Kmatrix = base_kernel(Gn, node_label, edge_label)
			
	# iterate each height
	for h in range(1, height + 1):
		all_set_compressed = {} # a dictionary mapping original labels to new ones in all graphs in this iteration
		num_of_labels_occured = 0 # number of the set of letters that occur before as node labels at least once in all graphs
		for G in Gn: # for each graph
			set_multisets = []		   
			for node in G.nodes(data = True):
				# Multiset-label determination.
				multiset = [ G.node[neighbors][node_label] for neighbors in G[node[0]] ]
				# sorting each multiset
				multiset.sort()
				multiset = node[1][node_label] + ''.join(multiset) # concatenate to a string and add the prefix 
				set_multisets.append(multiset)		  

			# label compression
			set_unique = list(set(set_multisets)) # set of unique multiset labels
			# a dictionary mapping original labels to new ones. 
			set_compressed = {}
			# if a label occured before, assign its former compressed label, else assign the number of labels occured + 1 as the compressed label 
			for value in set_unique:
				if value in all_set_compressed.keys():
					set_compressed.update({ value : all_set_compressed[value] })
				else:
					set_compressed.update({ value : str(num_of_labels_occured + 1) })
					num_of_labels_occured += 1

			all_set_compressed.update(set_compressed)
			
			# relabel nodes
			for node in G.nodes(data = True):
				node[1][node_label] = set_compressed[set_multisets[node[0]]]
				
		# Compute kernel with h iterations and add it to the final kernel
		Kmatrix += base_kernel(Gn, node_label, edge_label)
		
	return Kmatrix



import networkx as nx
import numpy as np
from copy import deepcopy
from enum import Enum, unique
#from itertools import product

# from tqdm import tqdm


#%%


def getSPLengths(G1):
	sp = nx.shortest_path(G1)
	distances = np.zeros((G1.number_of_nodes(), G1.number_of_nodes()))
	for i in sp.keys():
		for j in sp[i].keys():
			distances[i, j] = len(sp[i][j]) - 1
	return distances


def get_sp_graph(G, edge_weight=None):
	"""Transform graph G to its corresponding shortest-paths graph.

	Parameters
	----------
	G : NetworkX graph
		The graph to be transformed.
	edge_weight : string
		edge attribute corresponding to the edge weight.

	Return
	------
	S : NetworkX graph
		The shortest-paths graph corresponding to G.

	Notes
	------
	For an input graph G, its corresponding shortest-paths graph S contains the same set of nodes as G, while there exists an edge between all nodes in S which are connected by a walk in G. Every edge in S between two nodes is labeled by the shortest distance between these two nodes.

	References
	----------
	.. [1] Borgwardt KM, Kriegel HP. Shortest-path kernels on graphs. InData Mining, Fifth IEEE International Conference on 2005 Nov 27 (pp. 8-pp). IEEE.
	"""
	return floydTransformation(G, edge_weight=edge_weight)


def getSPGraph(G, edge_weight=None):
	"""Transform graph G to its corresponding shortest-paths graph.

	Parameters
	----------
	G : NetworkX graph
		The graph to be transformed.
	edge_weight : string
		edge attribute corresponding to the edge weight.

	Return
	------
	S : NetworkX graph
		The shortest-paths graph corresponding to G.

	Notes
	------
	For an input graph G, its corresponding shortest-paths graph S contains the same set of nodes as G, while there exists an edge between all nodes in S which are connected by a walk in G. Every edge in S between two nodes is labeled by the shortest distance between these two nodes.

	References
	----------
	.. [1] Borgwardt KM, Kriegel HP. Shortest-path kernels on graphs. InData Mining, Fifth IEEE International Conference on 2005 Nov 27 (pp. 8-pp). IEEE.
	"""
	# Raise deprecated warning:
	import warnings
	warnings.warn("getSPGraph is deprecated, use get_sp_graph instead.", DeprecationWarning)
	return floydTransformation(G, edge_weight=edge_weight)


def floydTransformation(G, edge_weight=None):
	"""Transform graph G to its corresponding shortest-paths graph using Floyd-transformation.

	Parameters
	----------
	G : NetworkX graph
		The graph to be tramsformed.
	edge_weight : string
		edge attribute corresponding to the edge weight. The default edge weight is bond_type.

	Return
	------
	S : NetworkX graph
		The shortest-paths graph corresponding to G.

	References
	----------
	.. [1] Borgwardt KM, Kriegel HP. Shortest-path kernels on graphs. InData Mining, Fifth IEEE International Conference on 2005 Nov 27 (pp. 8-pp). IEEE.
	"""
	spMatrix = nx.floyd_warshall_numpy(G, weight=edge_weight)
	S = nx.Graph()
	S.add_nodes_from(G.nodes(data=True))
	ns = list(G.nodes())
	for i in range(0, G.number_of_nodes()):
		for j in range(i + 1, G.number_of_nodes()):
			if spMatrix[i, j] != np.inf:
				S.add_edge(ns[i], ns[j], cost=spMatrix[i, j])
	return S


def get_shortest_paths(G, weight, directed):
	"""Get all shortest paths of a graph.

	Parameters
	----------
	G : NetworkX graphs
		The graphs whose paths are calculated.
	weight : string/None
		edge attribute used as weight to calculate the shortest path.
	directed: boolean
		Whether graph is directed.

	Return
	------
	sp : list of list
		List of shortest paths of the graph, where each path is represented by a list of nodes.
	"""
	from itertools import combinations
	sp = []
	for n1, n2 in combinations(G.nodes(), 2):
		try:
			spltemp = list(nx.all_shortest_paths(G, n1, n2, weight=weight))
		except nx.NetworkXNoPath:  # nodes not connected
			pass
		else:
			sp += spltemp
			# each edge walk is counted twice, starting from both its extreme nodes.
			if not directed:
				sp += [sptemp[::-1] for sptemp in spltemp]

	# add single nodes as length 0 paths.
	sp += [[n] for n in G.nodes()]
	return sp


def untotterTransformation(G, node_label, edge_label):
	"""Transform graph G according to Mahé et al.'s work to filter out tottering patterns of marginalized kernel and tree pattern kernel.

	Parameters
	----------
	G : NetworkX graph
		The graph to be tramsformed.
	node_label : string
		node attribute used as label. The default node label is 'atom'.
	edge_label : string
		edge attribute used as label. The default edge label is 'bond_type'.

	Return
	------
	gt : NetworkX graph
		The transformed graph corresponding to G.

	References
	----------
	.. [1] Pierre Mahé, Nobuhisa Ueda, Tatsuya Akutsu, Jean-Luc Perret, and Jean-Philippe Vert. Extensions of marginalized graph kernels. In Proceedings of the twenty-first international conference on Machine learning, page 70. ACM, 2004.
	"""
	# arrange all graphs in a list
	G = G.to_directed()
	gt = nx.Graph()
	gt.graph = G.graph
	gt.add_nodes_from(G.nodes(data=True))
	for edge in G.edges():
		gt.add_node(edge)
		gt.nodes[edge].update({node_label: G.nodes[edge[1]][node_label]})
		gt.add_edge(edge[0], edge)
		gt.edges[edge[0], edge].update({
			edge_label:
			G[edge[0]][edge[1]][edge_label]
		})
		for neighbor in G[edge[1]]:
			if neighbor != edge[0]:
				gt.add_edge(edge, (edge[1], neighbor))
				gt.edges[edge, (edge[1], neighbor)].update({
					edge_label:
					G[edge[1]][neighbor][edge_label]
				})
	# nx.draw_networkx(gt)
	# plt.show()

	# relabel nodes using consecutive integers for convenience of kernel calculation.
	gt = nx.convert_node_labels_to_integers(
		gt, first_label=0, label_attribute='label_orignal')
	return gt


def direct_product(G1, G2, node_label, edge_label):
	"""Return the direct/tensor product of directed graphs G1 and G2.

	Parameters
	----------
	G1, G2 : NetworkX graph
		The original graphs.
	node_label : string
		node attribute used as label. The default node label is 'atom'.
	edge_label : string
		edge attribute used as label. The default edge label is 'bond_type'.

	Return
	------
	gt : NetworkX graph
		The direct product graph of G1 and G2.

	Notes
	-----
	This method differs from networkx.tensor_product in that this method only adds nodes and edges in G1 and G2 that have the same labels to the direct product graph.

	References
	----------
	.. [1] Thomas Gärtner, Peter Flach, and Stefan Wrobel. On graph kernels: Hardness results and efficient alternatives. Learning Theory and Kernel Machines, pages 129–143, 2003.
	"""
	# arrange all graphs in a list
	from itertools import product
	# G = G.to_directed()
	gt = nx.DiGraph()
	# add nodes
	for u, v in product(G1, G2):
		if G1.nodes[u][node_label] == G2.nodes[v][node_label]:
			gt.add_node((u, v))
			gt.nodes[(u, v)].update({node_label: G1.nodes[u][node_label]})
	# add edges, faster for sparse graphs (no so many edges), which is the most case for now.
	for (u1, v1), (u2, v2) in product(G1.edges, G2.edges):
		if (u1, u2) in gt and (
				v1, v2
		) in gt and G1.edges[u1, v1][edge_label] == G2.edges[u2,
															 v2][edge_label]:
			gt.add_edge((u1, u2), (v1, v2))
			gt.edges[(u1, u2), (v1, v2)].update({
				edge_label:
				G1.edges[u1, v1][edge_label]
			})

	# # add edges, faster for dense graphs (a lot of edges, complete graph would be super).
	# for u, v in product(gt, gt):
	#	 if (u[0], v[0]) in G1.edges and (
	#			 u[1], v[1]
	#	 ) in G2.edges and G1.edges[u[0],
	#								v[0]][edge_label] == G2.edges[u[1],
	#															  v[1]][edge_label]:
	#		 gt.add_edge((u[0], u[1]), (v[0], v[1]))
	#		 gt.edges[(u[0], u[1]), (v[0], v[1])].update({
	#			 edge_label:
	#			 G1.edges[u[0], v[0]][edge_label]
	#		 })

	# relabel nodes using consecutive integers for convenience of kernel calculation.
	# gt = nx.convert_node_labels_to_integers(
	#	 gt, first_label=0, label_attribute='label_orignal')
	return gt


def direct_product_graph(G1, G2, node_labels, edge_labels):
	"""Return the direct/tensor product of directed graphs G1 and G2.

	Parameters
	----------
	G1, G2 : NetworkX graph
		The original graphs.
	node_labels : list
		A list of node attributes used as labels.
	edge_labels : list
		A list of edge attributes used as labels.

	Return
	------
	gt : NetworkX graph
		The direct product graph of G1 and G2.

	Notes
	-----
	This method differs from networkx.tensor_product in that this method only adds nodes and edges in G1 and G2 that have the same labels to the direct product graph.

	References
	----------
	.. [1] Thomas Gärtner, Peter Flach, and Stefan Wrobel. On graph kernels: Hardness results and efficient alternatives. Learning Theory and Kernel Machines, pages 129–143, 2003.
	"""
	# arrange all graphs in a list
	from itertools import product
	# G = G.to_directed()
	gt = nx.DiGraph()
	# add nodes
	for u, v in product(G1, G2):
		label1 = tuple(G1.nodes[u][nl] for nl in node_labels)
		label2 = tuple(G2.nodes[v][nl] for nl in node_labels)
		if label1 == label2:
			gt.add_node((u, v), node_label=label1)

	# add edges, faster for sparse graphs (no so many edges), which is the most case for now.
	for (u1, v1), (u2, v2) in product(G1.edges, G2.edges):
		if (u1, u2) in gt and (v1, v2) in gt:
			label1 = tuple(G1.edges[u1, v1][el] for el in edge_labels)
			label2 = tuple(G2.edges[u2, v2][el] for el in edge_labels)
			if label1 == label2:
				gt.add_edge((u1, u2), (v1, v2), edge_label=label1)


	# # add edges, faster for dense graphs (a lot of edges, complete graph would be super).
	# for u, v in product(gt, gt):
	#	 if (u[0], v[0]) in G1.edges and (
	#			 u[1], v[1]
	#	 ) in G2.edges and G1.edges[u[0],
	#								v[0]][edge_label] == G2.edges[u[1],
	#															  v[1]][edge_label]:
	#		 gt.add_edge((u[0], u[1]), (v[0], v[1]))
	#		 gt.edges[(u[0], u[1]), (v[0], v[1])].update({
	#			 edge_label:
	#			 G1.edges[u[0], v[0]][edge_label]
	#		 })

	# relabel nodes using consecutive integers for convenience of kernel calculation.
	# gt = nx.convert_node_labels_to_integers(
	#	 gt, first_label=0, label_attribute='label_orignal')
	return gt


def find_paths(G, source_node, length):
	"""Find all paths with a certain length those start from a source node.
	A recursive depth first search is applied.

	Parameters
	----------
	G : NetworkX graphs
		The graph in which paths are searched.
	source_node : integer
		The number of the node from where all paths start.
	length : integer
		The length of paths.

	Return
	------
	path : list of list
		List of paths retrieved, where each path is represented by a list of nodes.
	"""
	if length == 0:
		return [[source_node]]
	path = [[source_node] + path for neighbor in G[source_node] \
		for path in find_paths(G, neighbor, length - 1) if source_node not in path]
	return path


def find_all_paths(G, length, is_directed):
	"""Find all paths with a certain length in a graph. A recursive depth first
	search is applied.

	Parameters
	----------
	G : NetworkX graphs
		The graph in which paths are searched.
	length : integer
		The length of paths.

	Return
	------
	path : list of list
		List of paths retrieved, where each path is represented by a list of nodes.
	"""
	all_paths = []
	for node in G:
		all_paths.extend(find_paths(G, node, length))

	if not is_directed:
		# For each path, two presentations are retrieved from its two extremities.
		# Remove one of them.
		all_paths_r = [path[::-1] for path in all_paths]
		for idx, path in enumerate(all_paths[:-1]):
			for path2 in all_paths_r[idx+1::]:
				if path == path2:
					all_paths[idx] = []
					break
		all_paths = list(filter(lambda a: a != [], all_paths))

	return all_paths


# @todo: use it in ShortestPath.
def compute_vertex_kernels(g1, g2, node_kernels, node_labels=[], node_attrs=[]):
	"""Compute kernels between each pair of vertices in two graphs.

	Parameters
	----------
	g1, g2 : NetworkX graph
		The kernels bewteen pairs of vertices in these two graphs are computed.
	node_kernels : dict
		A dictionary of kernel functions for nodes, including 3 items: 'symb'
		for symbolic node labels, 'nsymb' for non-symbolic node labels, 'mix'
		for both labels. The first 2 functions take two node labels as
		parameters, and the 'mix' function takes 4 parameters, a symbolic and a
		non-symbolic label for each the two nodes. Each label is in form of 2-D
		dimension array (n_samples, n_features). Each function returns a number
		as the kernel value. Ignored when nodes are unlabeled. This argument
		is designated to conjugate gradient method and fixed-point iterations.
	node_labels : list, optional
		The list of the name strings of the node labels. The default is [].
	node_attrs : list, optional
		The list of the name strings of the node attributes. The default is [].

	Returns
	-------
	vk_dict : dict
		Vertex kernels keyed by vertices.

	Notes
	-----
	This function is used by ``gklearn.kernels.FixedPoint'' and
	``gklearn.kernels.StructuralSP''. The method is borrowed from FCSP [1].

	References
	----------
	.. [1] Lifan Xu, Wei Wang, M Alvarez, John Cavazos, and Dongping Zhang.
	Parallelization of shortest path graph kernels on multi-core cpus and gpus.
	Proceedings of the Programmability Issues for Heterogeneous Multicores
	(MultiProg), Vienna, Austria, 2014.
	"""
	vk_dict = {}  # shortest path matrices dict
	if len(node_labels) > 0:
		# node symb and non-synb labeled
		if len(node_attrs) > 0:
			kn = node_kernels['mix']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					n1_labels = [n1[1][nl] for nl in node_labels]
					n2_labels = [n2[1][nl] for nl in node_labels]
					# @TODO: reformat attrs during data processing a priori to save time.
					n1_attrs = np.array([n1[1][na] for na in node_attrs]).astype(float)
					n2_attrs = np.array([n2[1][na] for na in node_attrs]).astype(float)
					vk_dict[(n1[0], n2[0])] = kn(n1_labels, n2_labels, n1_attrs, n2_attrs)
		# node symb labeled
		else:
			kn = node_kernels['symb']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					n1_labels = [n1[1][nl] for nl in node_labels]
					n2_labels = [n2[1][nl] for nl in node_labels]
					vk_dict[(n1[0], n2[0])] = kn(n1_labels, n2_labels)
	else:
		# node non-synb labeled
		if len(node_attrs) > 0:
			kn = node_kernels['nsymb']
			for n1 in g1.nodes(data=True):
				for n2 in g2.nodes(data=True):
					n1_attrs = np.array([n1[1][na] for na in node_attrs]).astype(float)
					n2_attrs = np.array([n2[1][na] for na in node_attrs]).astype(float)
					vk_dict[(n1[0], n2[0])] = kn(n1_attrs, n2_attrs)
		# node unlabeled
		else:
			pass  # @todo: add edge weights.
# 			for e1 in g1.edges(data=True):
# 				for e2 in g2.edges(data=True):
# 					if e1[2]['cost'] == e2[2]['cost']:
# 						kernel += 1
# 			return kernel

	return vk_dict


#%%


def get_graph_kernel_by_name(name, node_labels=None, edge_labels=None, node_attrs=None, edge_attrs=None, ds_infos=None, kernel_options={}, **kwargs):
	if len(kwargs) != 0:
		kernel_options = kwargs

	if name == 'CommonWalk' or name == 'common walk':
		from gklearn.kernels import CommonWalk
		graph_kernel = CommonWalk(node_labels=node_labels,
								 edge_labels=edge_labels,
								 ds_infos=ds_infos,
								 **kernel_options)

	elif name == 'Marginalized' or name == 'marginalized':
		from gklearn.kernels import Marginalized
		graph_kernel = Marginalized(node_labels=node_labels,
								 edge_labels=edge_labels,
								 ds_infos=ds_infos,
								 **kernel_options)

	elif name == 'SylvesterEquation' or name == 'sylvester equation':
		from gklearn.kernels import SylvesterEquation
		graph_kernel = SylvesterEquation(
								  ds_infos=ds_infos,
								  **kernel_options)

	elif name == 'FixedPoint' or name == 'fixed point':
		from gklearn.kernels import FixedPoint
		graph_kernel = FixedPoint(node_labels=node_labels,
								  edge_labels=edge_labels,
								  node_attrs=node_attrs,
								  edge_attrs=edge_attrs,
								  ds_infos=ds_infos,
								  **kernel_options)

	elif name == 'ConjugateGradient' or name == 'conjugate gradient':
		from gklearn.kernels import ConjugateGradient
		graph_kernel = ConjugateGradient(node_labels=node_labels,
								  edge_labels=edge_labels,
								  node_attrs=node_attrs,
								  edge_attrs=edge_attrs,
								  ds_infos=ds_infos,
								  **kernel_options)

	elif name == 'SpectralDecomposition' or name == 'spectral decomposition':
		from gklearn.kernels import SpectralDecomposition
		graph_kernel = SpectralDecomposition(node_labels=node_labels,
								  edge_labels=edge_labels,
								  node_attrs=node_attrs,
								  edge_attrs=edge_attrs,
								  ds_infos=ds_infos,
								  **kernel_options)

	elif name == 'ShortestPath' or name == 'shortest path':
		from gklearn.kernels import ShortestPath
		graph_kernel = ShortestPath(node_labels=node_labels,
								 node_attrs=node_attrs,
								 ds_infos=ds_infos,
								 **kernel_options)

	elif name == 'StructuralSP' or name == 'structural shortest path':
		from gklearn.kernels import StructuralSP
		graph_kernel = StructuralSP(node_labels=node_labels,
								  edge_labels=edge_labels,
								  node_attrs=node_attrs,
								  edge_attrs=edge_attrs,
								  ds_infos=ds_infos,
								  **kernel_options)

	elif name == 'PathUpToH' or name == 'path up to length h':
		from gklearn.kernels import PathUpToH
		graph_kernel = PathUpToH(node_labels=node_labels,
							  edge_labels=edge_labels,
							  ds_infos=ds_infos,
							  **kernel_options)

	elif name == 'Treelet' or name == 'treelet':
		from gklearn.kernels import Treelet
		graph_kernel = Treelet(node_labels=node_labels,
							  edge_labels=edge_labels,
							  ds_infos=ds_infos,
							  **kernel_options)

	elif name == 'WLSubtree' or name == 'weisfeiler-lehman subtree':
		from gklearn.kernels import WLSubtree
		graph_kernel = WLSubtree(node_labels=node_labels,
							  edge_labels=edge_labels,
							  ds_infos=ds_infos,
							  **kernel_options)

	elif name == 'WeisfeilerLehman' or name == 'weisfeiler-lehman':
		from gklearn.kernels import WeisfeilerLehman
		graph_kernel = WeisfeilerLehman(node_labels=node_labels,
							  edge_labels=edge_labels,
							  ds_infos=ds_infos,
							  **kernel_options)
	else:
		raise Exception('The graph kernel given is not defined. Possible choices include: "StructuralSP", "ShortestPath", "PathUpToH", "Treelet", "WLSubtree", "WeisfeilerLehman".')

	return graph_kernel


def compute_gram_matrices_by_class(ds_name, kernel_options, save_results=True, dir_save='', irrelevant_labels=None, edge_required=False):
	import os
	from gklearn.utils import Dataset, split_dataset_by_target

	# 1. get dataset.
	print('1. getting dataset...')
	dataset_all = Dataset()
	dataset_all.load_predefined_dataset(ds_name)
	dataset_all.trim_dataset(edge_required=edge_required)
	if not irrelevant_labels is None:
		dataset_all.remove_labels(**irrelevant_labels)
# 	dataset_all.cut_graphs(range(0, 10))
	datasets = split_dataset_by_target(dataset_all)

	gram_matrix_unnorm_list = []
	run_time_list = []

	print('start generating preimage for each class of target...')
	for idx, dataset in enumerate(datasets):
		target = dataset.targets[0]
		print('\ntarget =', target, '\n')

		# 2. initialize graph kernel.
		print('2. initializing graph kernel and setting parameters...')
		graph_kernel = get_graph_kernel_by_name(kernel_options['name'],
										  node_labels=dataset.node_labels,
										  edge_labels=dataset.edge_labels,
										  node_attrs=dataset.node_attrs,
										  edge_attrs=dataset.edge_attrs,
										  ds_infos=dataset.get_dataset_infos(keys=['directed']),
										  kernel_options=kernel_options)

		# 3. compute gram matrix.
		print('3. computing gram matrix...')
		gram_matrix, run_time = graph_kernel.compute(dataset.graphs, **kernel_options)
		gram_matrix_unnorm = graph_kernel.gram_matrix_unnorm

		gram_matrix_unnorm_list.append(gram_matrix_unnorm)
		run_time_list.append(run_time)

	# 4. save results.
	print()
	print('4. saving results...')
	if save_results:
		os.makedirs(dir_save, exist_ok=True)
		np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm_list=gram_matrix_unnorm_list, run_time_list=run_time_list)

	print('\ncomplete.')


def normalize_gram_matrix(gram_matrix):
	diag = gram_matrix.diagonal().copy()
	old_settings = np.seterr(invalid='raise') # Catch FloatingPointError: invalid value encountered in sqrt.
	for i in range(len(gram_matrix)):
		for j in range(i, len(gram_matrix)):
			try:
				gram_matrix[i][j] /= np.sqrt(diag[i] * diag[j])
			except:
# 				rollback()
				np.seterr(**old_settings)
				raise
			else:
				gram_matrix[j][i] = gram_matrix[i][j]
	np.seterr(**old_settings)
	return gram_matrix


def compute_distance_matrix(gram_matrix):
	dis_mat = np.empty((len(gram_matrix), len(gram_matrix)))
	for i in range(len(gram_matrix)):
		for j in range(i, len(gram_matrix)):
			dis = gram_matrix[i, i] + gram_matrix[j, j] - 2 * gram_matrix[i, j]
			if dis < 0:
				if dis > -1e-10:
					dis = 0
				else:
					raise ValueError('The distance is negative.')
			dis_mat[i, j] = np.sqrt(dis)
			dis_mat[j, i] = dis_mat[i, j]
	dis_max = np.max(np.max(dis_mat))
	dis_min = np.min(np.min(dis_mat[dis_mat != 0]))
	dis_mean = np.mean(np.mean(dis_mat))
	return dis_mat, dis_max, dis_min, dis_mean


#%%


def graph_deepcopy(G):
	"""Deep copy a graph, including deep copy of all nodes, edges and
	attributes of the graph, nodes and edges.

	Note
	----
	- It is the same as the NetworkX function graph.copy(), as far as I know.

	- This function only supports Networkx.Graph and Networkx.DiGraph.
	"""
	# add graph attributes.
	labels = {}
	for k, v in G.graph.items():
		labels[k] = deepcopy(v)
	if G.is_directed():
		G_copy = nx.DiGraph(**labels)
	else:
		G_copy = nx.Graph(**labels)

	# add nodes
	for nd, attrs in G.nodes(data=True):
		labels = {}
		for k, v in attrs.items():
			labels[k] = deepcopy(v)
		G_copy.add_node(nd, **labels)

	# add edges.
	for nd1, nd2, attrs in G.edges(data=True):
		labels = {}
		for k, v in attrs.items():
			labels[k] = deepcopy(v)
		G_copy.add_edge(nd1, nd2, **labels)

	return G_copy


def graph_isIdentical(G1, G2):
	"""Check if two graphs are identical, including: same nodes, edges, node
	labels/attributes, edge labels/attributes.

	Notes
	-----
	1. The type of graphs has to be the same.

	2. Global/Graph attributes are neglected as they may contain names for graphs.
	"""
	# check nodes.
	nlist1 = [n for n in G1.nodes(data=True)]
	nlist2 = [n for n in G2.nodes(data=True)]
	if not nlist1 == nlist2:
		return False
	# check edges.
	elist1 = [n for n in G1.edges(data=True)]
	elist2 = [n for n in G2.edges(data=True)]
	if not elist1 == elist2:
		return False
	# check graph attributes.

	return True


def get_node_labels(Gn, node_label):
	"""Get node labels of dataset Gn.
	"""
	nl = set()
	for G in Gn:
		nl = nl | set(nx.get_node_attributes(G, node_label).values())
	return nl


def get_edge_labels(Gn, edge_label):
	"""Get edge labels of dataset Gn.
	"""
	el = set()
	for G in Gn:
		el = el | set(nx.get_edge_attributes(G, edge_label).values())
	return el


def get_mlti_dim_node_attrs(G, attr_names):
	attributes = []
	for nd, attrs in G.nodes(data=True):
		attributes.append(tuple(attrs[aname] for aname in attr_names))
	return attributes


def get_mlti_dim_edge_attrs(G, attr_names):
	attributes = []
	for ed, attrs in G.edges(data=True):
		attributes.append(tuple(attrs[aname] for aname in attr_names))
	return attributes


def nx_permute_nodes(G, random_state=None):
	"""Permute node indices in a NetworkX graph.

	Parameters
	----------
	G : TYPE
		DESCRIPTION.
	random_state : TYPE, optional
		DESCRIPTION. The default is None.

	Returns
	-------
	G_new : TYPE
		DESCRIPTION.

	Notes
	-----
	- This function only supports Networkx.Graph and Networkx.DiGraph.
	"""
	# @todo: relabel node with integers? (in case something went wrong...)
	# Add graph attributes.
	labels = {}
	for k, v in G.graph.items():
		labels[k] = deepcopy(v)
	if G.is_directed():
		G_new = nx.DiGraph(**labels)
	else:
		G_new = nx.Graph(**labels)

	# Create a random mapping old node indices <-> new indices.
	nb_nodes = nx.number_of_nodes(G)
	indices_orig = range(nb_nodes)
	idx_mapping = np.random.RandomState(seed=random_state).permutation(indices_orig)

	# Add nodes.
	nodes_orig = list(G.nodes)
	for i_orig in range(nb_nodes):
		i_new = idx_mapping[i_orig]
		labels = {}
		for k, v in G.nodes[nodes_orig[i_new]].items():
			labels[k] = deepcopy(v)
		G_new.add_node(nodes_orig[i_new], **labels)

	# Add edges.
	for nd1, nd2, attrs in G.edges(data=True):
		labels = {}
		for k, v in attrs.items():
			labels[k] = deepcopy(v)
		G_new.add_edge(nd1, nd2, **labels)


# 	# create a random mapping old label -> new label
# 	node_mapping = dict(zip(G.nodes(), np.random.RandomState(seed=random_state).permutation(G.nodes())))
# 	# build a new graph
# 	G_new = nx.relabel_nodes(G, node_mapping)

	return G_new


#%%


def dummy_node():
	"""
	/*!
	 * @brief Returns a dummy node.
	 * @return ID of dummy node.
	 */
	"""
	return np.inf # @todo: in GEDLIB, this is the max - 1 rather than max, I don't know why.


def undefined_node():
	"""
	/*!
	 * @brief Returns an undefined node.
	 * @return ID of undefined node.
	 */

	"""
	return np.inf


def dummy_edge():
	"""
	/*!
	 * @brief Returns a dummy edge.
	 * @return ID of dummy edge.
	 */

	"""
	return np.inf


@unique
class SpecialLabel(Enum):
	"""can be used to define special labels.
	"""
	DUMMY = 1 # The dummy label.
	# DUMMY = auto # enum.auto does not exist in Python 3.5.


#%%


def check_json_serializable(
		obj,
		deep: bool = False,
) -> bool:
	"""Check if an object is JSON serializable.

	Parameters
	----------
	obj : object
		The object to be checked.

	deep : bool, optional
		Whether to check the object recursively when `obj` is iterable.
		The default is False.

	Returns
	-------
	bool
		True if the object is JSON serializable, False otherwise.
	"""
	import json
	try:
		json.dumps(obj)
	except TypeError:
		return False
	else:
		if deep and hasattr(obj, '__iter__'):
			for item in obj:
				if not check_json_serializable(item, deep=True):
					return False
		return True


def is_basic_python_type(
		obj,
		type_list: list = None,
		deep: bool = False,
) -> bool:
	"""Check if an object is a basic type in Python.

	Parameters
	----------
	obj : object
		The object to be checked.

	type_list : list, optional
		The list of basic types in Python. The default is None, which means
		the default basic types are used. The default basic types include
		`int`, `float`, `complex`, `str`, `bool`, `NoneType`, `list`,
		`tuple`, `dict`, `set`, `frozenset`, `range`, `slice`.

	deep : bool, optional
		Whether to check the object recursively when `obj` is iterable.
		The default is False.

	Returns
	-------
	bool
		True if the object is a basic type in Python, False otherwise.
	"""
	if type_list is None:
		type_list = [
			int, float, complex, str, bool, type(None), list, tuple, dict,
			set, frozenset, range, slice
		]
	if not hasattr(obj, '__iter__') or isinstance(obj, (str, bytes)):
		return type(obj) in type_list
	else:
		if deep:
			for item in obj:
				if not is_basic_python_type(item, type_list=type_list, deep=True):
					return False
			return True
		else:
			return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 18 15:52:35 2020

@author: ljia
"""
import numpy as np
import time
import networkx as nx


class GEDMethod(object):
	
	
	def __init__(self, ged_data):
		self._initialized = False
		self._ged_data = ged_data
		self._options = None
		self._lower_bound = 0
		self._upper_bound = np.inf
		self._node_map = [0, 0] # @todo
		self._runtime = None
		self._init_time = None
		
	
	def init(self):
		"""Initializes the method with options specified by set_options().
		"""
		start = time.time()
		self._ged_init()
		end = time.time()
		self._init_time = end - start
		self._initialized = True
		
		
	def set_options(self, options):
		"""
	/*!
	 * @brief Sets the options of the method.
	 * @param[in] options String of the form <tt>[--@<option@> @<arg@>] [...]</tt>, where @p option contains neither spaces nor single quotes,
	 * and @p arg contains neither spaces nor single quotes or is of the form <tt>'[--@<sub-option@> @<sub-arg@>] [...]'</tt>,
	 * where both @p sub-option and @p sub-arg contain neither spaces nor single quotes.
	 */
		"""
		self._ged_set_default_options()
		for key, val in options.items():
			if not self._ged_parse_option(key, val):
				raise Exception('Invalid option "', key, '". Usage: options = "' + self._ged_valid_options_string() + '".') # @todo: not implemented.
		self._initialized = False
		
		
	def run(self, g_id, h_id):
		"""
	/*!
	 * @brief Runs the method with options specified by set_options().
	 * @param[in] g_id ID of input graph.
	 * @param[in] h_id ID of input graph.
	 */
		"""
		start = time.time()
		result = self.run_as_util(self._ged_data._graphs[g_id], self._ged_data._graphs[h_id])
		end = time.time()
		self._lower_bound = result['lower_bound']
		self._upper_bound = result['upper_bound']
		if len(result['node_maps']) > 0:
			self._node_map = result['node_maps'][0]
		self._runtime = end - start
		
		
	def run_as_util(self, g, h):
		"""
	/*!
	 * @brief Runs the method with options specified by set_options().
	 * @param[in] g Input graph.
	 * @param[in] h Input graph.
	 * @param[out] result Result variable.
	 */
		"""
		# Compute optimal solution and return if at least one of the two graphs is empty.
		if nx.number_of_nodes(g) == 0 or nx.number_of_nodes(h) == 0:
			print('This is not implemented.')
			pass # @todo:
			
		# Run the method.
		return self._ged_run(g, h)
		
		
	def get_upper_bound(self):
		"""
	/*!
	 * @brief Returns an upper bound.
	 * @return Upper bound for graph edit distance provided by last call to run() or -1 if the method does not yield an upper bound.
	 */
		"""
		return self._upper_bound
		
		
	def get_lower_bound(self):
		"""
	/*!
	 * @brief Returns a lower bound.
	 * @return Lower bound for graph edit distance provided by last call to run() or -1 if the method does not yield a lower bound.
	 */
		"""
		return self._lower_bound
		
		
	def get_runtime(self):
		"""
	/*!
	 * @brief Returns the runtime.
	 * @return Runtime of last call to run() in seconds.
	 */
		"""
		return self._runtime
	

	def get_init_time(self):
		"""
	/*!
	 * @brief Returns the initialization time.
	 * @return Runtime of last call to init() in seconds.
	 */
		"""		
		return self._init_time


	def get_node_map(self):
		"""
	/*!
	 * @brief Returns a graph matching.
	 * @return Constant reference to graph matching provided by last call to run() or to an empty matching if the method does not yield a matching.
	 */
		"""
		return self._node_map
		
		
	def _ged_init(self):
		"""
	/*!
	 * @brief Initializes the method.
	 * @note Must be overridden by derived classes that require initialization.
	 */
		"""
		pass
			
			
	def _ged_parse_option(self, option, arg):
		"""
	/*!
	 * @brief Parses one option.
	 * @param[in] option The name of the option.
	 * @param[in] arg The argument of the option.
	 * @return Boolean @p true if @p option is a valid option name for the method and @p false otherwise.
	 * @note Must be overridden by derived classes that have options.
	 */
		"""
		return False
	
	
	def _ged_run(self, g, h):
		"""
	/*!
	 * @brief Runs the method with options specified by set_options().
	 * @param[in] g Input graph.
	 * @param[in] h Input graph.
	 * @param[out] result Result variable.
	 * @note Must be overridden by derived classes.
	 */
		"""
		return {}
		
	
	
	def _ged_valid_options_string(self):
		"""
	/*!
	 * @brief Returns string of all valid options.
	 * @return String of the form <tt>[--@<option@> @<arg@>] [...]</tt>.
	 * @note Must be overridden by derived classes that have options.
	 */
		"""
		return ''
		
		
	def _ged_set_default_options(self):
		"""
	/*!
	 * @brief Sets all options to default values.
	 * @note Must be overridden by derived classes that have options.
	 */
		"""
		pass
		


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 24 14:25:57 2022

@author: ljia
"""
from ._split import BaseCrossValidatorWithValid
# from ._split import BaseShuffleSplit
from ._split import KFoldWithValid
# from ._split import GroupKFold
# from ._split import StratifiedKFoldWithValid
# from ._split import TimeSeriesSplit
# from ._split import LeaveOneGroupOut
# from ._split import LeaveOneOut
# from ._split import LeavePGroupsOut
# from ._split import LeavePOut
from ._split import RepeatedKFoldWithValid
# from ._split import RepeatedStratifiedKFold
# from ._split import ShuffleSplit
# from ._split import GroupShuffleSplit
# from ._split import StratifiedShuffleSplit
# from ._split import StratifiedGroupKFold
# from ._split import PredefinedSplit


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 17 12:02:36 2020

@author: ljia
"""
import numpy as np
import networkx as nx
from gklearn.ged.env import Options, OptionsStringMap
from gklearn.ged.env import GEDData


class GEDEnv(object):

	
	def __init__(self):
		self._initialized = False
		self._new_graph_ids = []
		self._ged_data = GEDData()
		# Variables needed for approximating ged_instance_.
		self._lower_bounds = {}
		self._upper_bounds = {}
		self._runtimes = {}
		self._node_maps = {}
		self._original_to_internal_node_ids = []
		self._internal_to_original_node_ids = []
		self._ged_method = None
	
		
	def set_edit_cost(self, edit_cost, edit_cost_constants=[]):
		"""
	/*!
	 * @brief Sets the edit costs to one of the predefined edit costs.
	 * @param[in] edit_costs Select one of the predefined edit costs.
	 * @param[in] edit_cost_constants Constants passed to the constructor of the edit cost class selected by @p edit_costs.
	 */
		"""
		self._ged_data._set_edit_cost(edit_cost, edit_cost_constants)
		
		
	def add_graph(self, graph_name='', graph_class=''):
		"""
	/*!
	 * @brief Adds a new uninitialized graph to the environment. Call init() after calling this method.
	 * @param[in] graph_name The name of the added graph. Empty if not specified.
	 * @param[in] graph_class The class of the added graph. Empty if not specified.
	 * @return The ID of the newly added graph.
	 */
		"""
		# @todo: graphs are not uninitialized.
		self._initialized = False
		graph_id = self._ged_data._num_graphs_without_shuffled_copies
		self._ged_data._num_graphs_without_shuffled_copies += 1
		self._new_graph_ids.append(graph_id)
		self._ged_data._graphs.append(nx.Graph())
		self._ged_data._graph_names.append(graph_name)
		self._ged_data._graph_classes.append(graph_class)
		self._original_to_internal_node_ids.append({})
		self._internal_to_original_node_ids.append({})
		self._ged_data._strings_to_internal_node_ids.append({})
		self._ged_data._internal_node_ids_to_strings.append({})
		return graph_id
	
	
	def clear_graph(self, graph_id):
		"""
	/*!
	 * @brief Clears and de-initializes a graph that has previously been added to the environment. Call init() after calling this method.
	 * @param[in] graph_id ID of graph that has to be cleared.
	 */
		"""
		if graph_id > self._ged_data.num_graphs_without_shuffled_copies():
			raise Exception('The graph', self.get_graph_name(graph_id), 'has not been added to the environment.')
		self._ged_data._graphs[graph_id].clear()
		self._original_to_internal_node_ids[graph_id].clear()
		self._internal_to_original_node_ids[graph_id].clear()
		self._ged_data._strings_to_internal_node_ids[graph_id].clear()
		self._ged_data._internal_node_ids_to_strings[graph_id].clear()
		self._initialized = False
	
	
	def add_node(self, graph_id, node_id, node_label):
		"""
	/*!
	 * @brief Adds a labeled node.
	 * @param[in] graph_id ID of graph that has been added to the environment.
	 * @param[in] node_id The user-specific ID of the vertex that has to be added.
	 * @param[in] node_label The label of the vertex that has to be added. Set to ged::NoLabel() if template parameter @p UserNodeLabel equals ged::NoLabel.
	 */
		"""
		# @todo: check ids.
		self._initialized = False
		internal_node_id = nx.number_of_nodes(self._ged_data._graphs[graph_id])
		self._ged_data._graphs[graph_id].add_node(internal_node_id, label=node_label)
		self._original_to_internal_node_ids[graph_id][node_id] = internal_node_id
		self._internal_to_original_node_ids[graph_id][internal_node_id] = node_id
		self._ged_data._strings_to_internal_node_ids[graph_id][str(node_id)] = internal_node_id
		self._ged_data._internal_node_ids_to_strings[graph_id][internal_node_id] = str(node_id)
		self._ged_data._node_label_to_id(node_label)
		label_id = self._ged_data._node_label_to_id(node_label)
		# @todo: ged_data_.graphs_[graph_id].set_label
		
		
	def add_edge(self, graph_id, nd_from, nd_to, edge_label, ignore_duplicates=True):
		"""
	/*!
	 * @brief Adds a labeled edge.
	 * @param[in] graph_id ID of graph that has been added to the environment.
	 * @param[in] tail The user-specific ID of the tail of the edge that has to be added.
	 * @param[in] head The user-specific ID of the head of the edge that has to be added.
	 * @param[in] edge_label The label of the vertex that has to be added. Set to ged::NoLabel() if template parameter @p UserEdgeLabel equals ged::NoLabel.
	 * @param[in] ignore_duplicates If @p true, duplicate edges are ignores. Otherwise, an exception is thrown if an existing edge is added to the graph.
	 */
		"""
		# @todo: check everything.
		self._initialized = False
		# @todo: check ignore_duplicates.
		self._ged_data._graphs[graph_id].add_edge(self._original_to_internal_node_ids[graph_id][nd_from], self._original_to_internal_node_ids[graph_id][nd_to], label=edge_label)
		label_id = self._ged_data._edge_label_to_id(edge_label)
		# @todo: ged_data_.graphs_[graph_id].set_label
		
	
	def add_nx_graph(self, g, classe, ignore_duplicates=True) :
		"""
			Add a Graph (made by networkx) on the environment. Be careful to respect the same format as GXL graphs for labelling nodes and edges. 
	
			:param g: The graph to add (networkx graph)
			:param ignore_duplicates: If True, duplicate edges are ignored, otherwise it's raise an error if an existing edge is added. True by default
			:type g: networkx.graph
			:type ignore_duplicates: bool
			:return: The ID of the newly added graphe
			:rtype: size_t
	
			.. note:: The NX graph must respect the GXL structure. Please see how a GXL graph is construct.  
			
		"""
		graph_id = self.add_graph(g.name, classe) # check if the graph name already exists.
		for node in g.nodes: # @todo: if the keys of labels include int and str at the same time.
			self.add_node(graph_id, node, tuple(sorted(g.nodes[node].items(), key=lambda kv: kv[0])))
		for edge in g.edges:
			self.add_edge(graph_id, edge[0], edge[1], tuple(sorted(g.edges[(edge[0], edge[1])].items(), key=lambda kv: kv[0])), ignore_duplicates)
		return graph_id
	
	
	def load_nx_graph(self, nx_graph, graph_id, graph_name='', graph_class=''):
		"""
		Loads NetworkX Graph into the GED environment.

		Parameters
		----------
		nx_graph : NetworkX Graph object
			The graph that should be loaded.
			
		graph_id : int or None
			The ID of a graph contained the environment (overwrite existing graph) or add new graph if `None`.
																							
		graph_name : string, optional
			The name of newly added graph. The default is ''. Has no effect unless `graph_id` equals `None`.
			
		graph_class : string, optional
			The class of newly added graph. The default is ''. Has no effect unless `graph_id` equals `None`.

		Returns
		-------
		int
			The ID of the newly loaded graph.
		"""
		if graph_id is None: # @todo: undefined.
			graph_id = self.add_graph(graph_name, graph_class)
		else:
			self.clear_graph(graph_id)
		for node in nx_graph.nodes:
			self.add_node(graph_id, node, tuple(sorted(nx_graph.nodes[node].items(), key=lambda kv: kv[0])))
		for edge in nx_graph.edges:
			self.add_edge(graph_id, edge[0], edge[1], tuple(sorted(nx_graph.edges[(edge[0], edge[1])].items(), key=lambda kv: kv[0])))
		return graph_id
	
	
	def init(self, init_type=Options.InitType.EAGER_WITHOUT_SHUFFLED_COPIES, print_to_stdout=False):
		if isinstance(init_type, str):
			init_type = OptionsStringMap.InitType[init_type]
			
		# Throw an exception if no edit costs have been selected.
		if self._ged_data._edit_cost is None:
			raise Exception('No edit costs have been selected. Call set_edit_cost() before calling init().')
			
		# Return if the environment is initialized.
		if self._initialized:
			return
		
		# Set initialization type.
		self._ged_data._init_type = init_type
		
		# @todo: Construct shuffled graph copies if necessary.
		
		# Re-initialize adjacency matrices (also previously initialized graphs must be re-initialized because of possible re-allocation).
		# @todo: setup_adjacency_matrix, don't know if neccessary.
		self._ged_data._max_num_nodes = np.max([nx.number_of_nodes(g) for g in self._ged_data._graphs])
		self._ged_data._max_num_edges = np.max([nx.number_of_edges(g) for g in self._ged_data._graphs])
			
		# Initialize cost matrices if necessary.
		if self._ged_data._eager_init():
			pass # @todo: init_cost_matrices_: 1. Update node cost matrix if new node labels have been added to the environment; 2. Update edge cost matrix if new edge labels have been added to the environment.
			
		# Mark environment as initialized.
		self._initialized = True
		self._new_graph_ids.clear()
		
		
	def is_initialized(self):
		"""
	/*!
	 * @brief Check if the environment is initialized.
	 * @return True if the environment is initialized.
	 */
		"""
		return self._initialized
	
	
	def get_init_type(self):
		"""
	/*!
	 * @brief Returns the initialization type of the last initialization.
	 * @return Initialization type.
	 */
		"""
		return self._ged_data._init_type
	
	
	def set_label_costs(self, node_label_costs=None, edge_label_costs=None):
		"""Set the costs between labels. 
		"""
		if node_label_costs is not None:
			self._ged_data._node_label_costs = node_label_costs
		if edge_label_costs is not None:
			self._ged_data._edge_label_costs = edge_label_costs
		
		
	def set_method(self, method, options=''):
		"""
	/*!
	 * @brief Sets the GEDMethod to be used by run_method().
	 * @param[in] method Select the method that is to be used.
	 * @param[in] options An options string of the form @"[--@<option@> @<arg@>] [...]@" passed to the selected method.
	 */
		"""
		del self._ged_method
		
		if isinstance(method, str):
			method = OptionsStringMap.GEDMethod[method]

		if method == Options.GEDMethod.BRANCH:
			self._ged_method = Branch(self._ged_data)
		elif method == Options.GEDMethod.BRANCH_FAST:
			self._ged_method = BranchFast(self._ged_data)
		elif method == Options.GEDMethod.BRANCH_FAST:
			self._ged_method = BranchFast(self._ged_data)	
		elif method == Options.GEDMethod.BRANCH_TIGHT:
			self._ged_method = BranchTight(self._ged_data)	
		elif method == Options.GEDMethod.BRANCH_UNIFORM:
			self._ged_method = BranchUniform(self._ged_data)	
		elif method == Options.GEDMethod.BRANCH_COMPACT:
			self._ged_method = BranchCompact(self._ged_data)	
		elif method == Options.GEDMethod.PARTITION:
			self._ged_method = Partition(self._ged_data)	
		elif method == Options.GEDMethod.HYBRID:
			self._ged_method = Hybrid(self._ged_data)	
		elif method == Options.GEDMethod.RING:
			self._ged_method = Ring(self._ged_data)	
		elif method == Options.GEDMethod.ANCHOR_AWARE_GED:
			self._ged_method = AnchorAwareGED(self._ged_data)	
		elif method == Options.GEDMethod.WALKS:
			self._ged_method = Walks(self._ged_data)	
		elif method == Options.GEDMethod.IPFP:
			self._ged_method = IPFP(self._ged_data)	
		elif method == Options.GEDMethod.BIPARTITE:
			from gklearn.ged.methods import Bipartite
			self._ged_method = Bipartite(self._ged_data)	
		elif method == Options.GEDMethod.SUBGRAPH:
			self._ged_method = Subgraph(self._ged_data)	
		elif method == Options.GEDMethod.NODE:
			self._ged_method = Node(self._ged_data)	
		elif method == Options.GEDMethod.RING_ML:
			self._ged_method = RingML(self._ged_data)	
		elif method == Options.GEDMethod.BIPARTITE_ML:
			self._ged_method = BipartiteML(self._ged_data)	
		elif method == Options.GEDMethod.REFINE:
			self._ged_method = Refine(self._ged_data)	
		elif method == Options.GEDMethod.BP_BEAM:
			self._ged_method = BPBeam(self._ged_data)	
		elif method == Options.GEDMethod.SIMULATED_ANNEALING:
			self._ged_method = SimulatedAnnealing(self._ged_data)	
		elif method == Options.GEDMethod.HED:
			self._ged_method = HED(self._ged_data)	
		elif method == Options.GEDMethod.STAR:
			self._ged_method = STAR(self._ged_data)	
		# #ifdef GUROBI
		elif method == Options.GEDMethod.F1:
			self._ged_method = F1(self._ged_data)	
		elif method == Options.GEDMethod.F2:
			self._ged_method = F2(self._ged_data)	
		elif method == Options.GEDMethod.COMPACT_MIP:
			self._ged_method = CompactMIP(self._ged_data)	
		elif method == Options.GEDMethod.BLP_NO_EDGE_LABELS:
			self._ged_method = BLPNoEdgeLabels(self._ged_data)	

		self._ged_method.set_options(options)
		
		
	def run_method(self, g_id, h_id):
		"""
	/*!
	 * @brief Runs the GED method specified by call to set_method() between the graphs with IDs @p g_id and @p h_id.
	 * @param[in] g_id ID of an input graph that has been added to the environment.
	 * @param[in] h_id ID of an input graph that has been added to the environment.
	 */
		"""
		if g_id >= self._ged_data.num_graphs():
			raise Exception('The graph with ID', str(g_id), 'has not been added to the environment.')
		if h_id >= self._ged_data.num_graphs():
			raise Exception('The graph with ID', str(h_id), 'has not been added to the environment.')
		if not self._initialized:
			raise Exception('The environment is uninitialized. Call init() after adding all graphs to the environment.')
		if self._ged_method is None:
			raise Exception('No method has been set. Call set_method() before calling run().')
		
		# Call selected GEDMethod and store results.
		if self._ged_data.shuffled_graph_copies_available() and (g_id == h_id):
			self._ged_method.run(g_id, self._ged_data.id_shuffled_graph_copy(h_id)) # @todo: why shuffle?
		else:
			self._ged_method.run(g_id, h_id)
		self._lower_bounds[(g_id, h_id)] = self._ged_method.get_lower_bound()
		self._upper_bounds[(g_id, h_id)] = self._ged_method.get_upper_bound()
		self._runtimes[(g_id, h_id)] = self._ged_method.get_runtime()
		self._node_maps[(g_id, h_id)] = self._ged_method.get_node_map()
		
		
	def init_method(self):
		"""Initializes the method specified by call to set_method().
		"""
		if not self._initialized:
			raise Exception('The environment is uninitialized. Call init() before calling init_method().')
		if self._ged_method is None:
			raise Exception('No method has been set. Call set_method() before calling init_method().')
		self._ged_method.init()
		
		
	def get_num_node_labels(self):
		"""
	/*!
	 * @brief Returns the number of node labels.
	 * @return Number of pairwise different node labels contained in the environment.
	 * @note If @p 1 is returned, the nodes are unlabeled.
	 */
		"""
		return len(self._ged_data._node_labels)
	
	
	def get_all_node_labels(self):
		"""
	/*!
	 * @brief Returns the list of all node labels.
	 * @return List of pairwise different node labels contained in the environment.
	 * @note If @p 1 is returned, the nodes are unlabeled.
	 */
		"""
		return self._ged_data._node_labels
	
	
	def get_node_label(self, label_id, to_dict=True):
		"""
	/*!
	 * @brief Returns node label.
	 * @param[in] label_id ID of node label that should be returned. Must be between 1 and num_node_labels().
	 * @return Node label for selected label ID.
	 */
		"""
		if label_id < 1 or label_id > self.get_num_node_labels():
			raise Exception('The environment does not contain a node label with ID', str(label_id), '.')
		if to_dict:
			return dict(self._ged_data._node_labels[label_id - 1])
		return self._ged_data._node_labels[label_id - 1]
		
		
	def get_num_edge_labels(self):
		"""
	/*!
	 * @brief Returns the number of edge labels.
	 * @return Number of pairwise different edge labels contained in the environment.
	 * @note If @p 1 is returned, the edges are unlabeled.
	 */
		"""
		return len(self._ged_data._edge_labels)
	
	
	def get_all_edge_labels(self):
		"""
	/*!
	 * @brief Returns the list of all edge labels.
	 * @return List of pairwise different edge labels contained in the environment.
	 * @note If @p 1 is returned, the edges are unlabeled.
	 */
		"""
		return self._ged_data._edge_labels
	
		
	def get_edge_label(self, label_id, to_dict=True):
		"""
	/*!
	 * @brief Returns edge label.
	 * @param[in] label_id ID of edge label that should be returned. Must be between 1 and num_node_labels().
	 * @return Edge label for selected label ID.
	 */
		"""
		if label_id < 1 or label_id > self.get_num_edge_labels():
			raise Exception('The environment does not contain an edge label with ID', str(label_id), '.')
		if to_dict:
			return dict(self._ged_data._edge_labels[label_id - 1])
		return self._ged_data._edge_labels[label_id - 1]
	
		
	def get_upper_bound(self, g_id, h_id):
		"""
	/*!
	 * @brief Returns upper bound for edit distance between the input graphs.
	 * @param[in] g_id ID of an input graph that has been added to the environment.
	 * @param[in] h_id ID of an input graph that has been added to the environment.
	 * @return Upper bound computed by the last call to run_method() with arguments @p g_id and @p h_id.
	 */
		"""
		if (g_id, h_id) not in self._upper_bounds:
			raise Exception('Call run(' + str(g_id) + ',' + str(h_id) + ') before calling get_upper_bound(' + str(g_id) + ',' + str(h_id) + ').')
		return self._upper_bounds[(g_id, h_id)]
		
		
	def get_lower_bound(self, g_id, h_id):
		"""
	/*!
	 * @brief Returns lower bound for edit distance between the input graphs.
	 * @param[in] g_id ID of an input graph that has been added to the environment.
	 * @param[in] h_id ID of an input graph that has been added to the environment.
	 * @return Lower bound computed by the last call to run_method() with arguments @p g_id and @p h_id.
	 */
		"""
		if (g_id, h_id) not in self._lower_bounds:
			raise Exception('Call run(' + str(g_id) + ',' + str(h_id) + ') before calling get_lower_bound(' + str(g_id) + ',' + str(h_id) + ').')
		return self._lower_bounds[(g_id, h_id)]
		
		
	def get_runtime(self, g_id, h_id):
		"""
	/*!
	 * @brief Returns runtime.
	 * @param[in] g_id ID of an input graph that has been added to the environment.
	 * @param[in] h_id ID of an input graph that has been added to the environment.
	 * @return Runtime of last call to run_method() with arguments @p g_id and @p h_id.
	 */
		"""
		if (g_id, h_id) not in self._runtimes:
			raise Exception('Call run(' + str(g_id) + ',' + str(h_id) + ') before calling get_runtime(' + str(g_id) + ',' + str(h_id) + ').')
		return self._runtimes[(g_id, h_id)]
	

	def get_init_time(self):
		"""
	/*!
	 * @brief Returns initialization time.
	 * @return Runtime of the last call to init_method().
	 */
		"""		
		return self._ged_method.get_init_time()


	def get_node_map(self, g_id, h_id):
		"""
	/*!
	 * @brief Returns node map between the input graphs.
	 * @param[in] g_id ID of an input graph that has been added to the environment.
	 * @param[in] h_id ID of an input graph that has been added to the environment.
	 * @return Node map computed by the last call to run_method() with arguments @p g_id and @p h_id.
	 */
		"""
		if (g_id, h_id) not in self._node_maps:
			raise Exception('Call run(' + str(g_id) + ',' + str(h_id) + ') before calling get_node_map(' + str(g_id) + ',' + str(h_id) + ').')
		return self._node_maps[(g_id, h_id)]
	

	def get_forward_map(self, g_id, h_id) :
		"""
			Returns the forward map (or the half of the adjacence matrix) between nodes of the two indicated graphs. 
	
			:param g: The Id of the first compared graph 
			:param h: The Id of the second compared graph
			:type g: size_t
			:type h: size_t
			:return: The forward map to the adjacence matrix between nodes of the two graphs
			:rtype: list[npy_uint32]
			
			.. seealso:: run_method(), get_upper_bound(), get_lower_bound(), get_backward_map(), get_runtime(), quasimetric_cost(), get_node_map(), get_assignment_matrix()
			.. warning:: run_method() between the same two graph must be called before this function. 
			.. note:: I don't know how to connect the two map to reconstruct the adjacence matrix. Please come back when I know how it's work ! 
		"""
		return self.get_node_map(g_id, h_id).forward_map
	
	
	def get_backward_map(self, g_id, h_id) :
		"""
			Returns the backward map (or the half of the adjacence matrix) between nodes of the two indicated graphs. 
	
			:param g: The Id of the first compared graph 
			:param h: The Id of the second compared graph
			:type g: size_t
			:type h: size_t
			:return: The backward map to the adjacence matrix between nodes of the two graphs
			:rtype: list[npy_uint32]
			
			.. seealso:: run_method(), get_upper_bound(), get_lower_bound(),  get_forward_map(), get_runtime(), quasimetric_cost(), get_node_map(), get_assignment_matrix()
			.. warning:: run_method() between the same two graph must be called before this function. 
			.. note:: I don't know how to connect the two map to reconstruct the adjacence matrix. Please come back when I know how it's work ! 
		"""
		return self.get_node_map(g_id, h_id).backward_map
	
	
	def compute_induced_cost(self, g_id, h_id, node_map):
		"""
	/*!
	 * @brief Computes the edit cost between two graphs induced by a node map.
	 * @param[in] g_id ID of input graph.
	 * @param[in] h_id ID of input graph.
	 * @param[in,out] node_map Node map whose induced edit cost is to be computed.
	 */
		"""
		self._ged_data.compute_induced_cost(self._ged_data._graphs[g_id], self._ged_data._graphs[h_id], node_map)
	
	
	def get_nx_graph(self, graph_id):
		"""
	 * @brief Returns NetworkX.Graph() representation.
	 * @param[in] graph_id ID of the selected graph.
		"""
		graph = nx.Graph() # @todo: add graph attributes.
		graph.graph['id'] = graph_id
		
		nb_nodes = self.get_graph_num_nodes(graph_id)
		original_node_ids = self.get_original_node_ids(graph_id)
		node_labels = self.get_graph_node_labels(graph_id, to_dict=True)
		graph.graph['original_node_ids'] = original_node_ids		
		
		for node_id in range(0, nb_nodes):
			graph.add_node(node_id, **node_labels[node_id])
			
		edges = self.get_graph_edges(graph_id, to_dict=True)
		for (head, tail), labels in edges.items():
			graph.add_edge(head, tail, **labels)

		return graph
	
	
	def get_graph_node_labels(self, graph_id, to_dict=True):
		"""
			Searchs and returns all the labels of nodes on a graph, selected by its ID. 
	
			:param graph_id: The ID of the wanted graph
			:type graph_id: size_t
			:return: The list of nodes' labels on the selected graph
			:rtype: list[dict{string : string}]
			
			.. seealso:: get_graph_internal_id(), get_graph_num_nodes(), get_graph_num_edges(), get_original_node_ids(), get_graph_edges(), get_graph_adjacence_matrix()
			.. note:: These functions allow to collect all the graph's informations.
		"""
		graph = self._ged_data.graph(graph_id)
		node_labels = []
		for n in graph.nodes():
			node_labels.append(graph.nodes[n]['label'])
		if to_dict:
			return [dict(i) for i in node_labels]
		return node_labels
	
	
	def get_graph_edges(self, graph_id, to_dict=True):
		"""
			Searchs and returns all the edges on a graph, selected by its ID. 
	
			:param graph_id: The ID of the wanted graph
			:type graph_id: size_t
			:return: The list of edges on the selected graph
			:rtype: dict{tuple(size_t, size_t) : dict{string : string}}
			
			.. seealso::get_graph_internal_id(), get_graph_num_nodes(), get_graph_num_edges(), get_original_node_ids(), get_graph_node_labels(), get_graph_adjacence_matrix()
			.. note:: These functions allow to collect all the graph's informations.
		"""
		graph = self._ged_data.graph(graph_id)		
		if to_dict:
			edges = {}		
			for n1, n2, attr in graph.edges(data=True):
				edges[(n1, n2)] = dict(attr['label'])
			return edges
		return {(n1, n2): attr['label'] for n1, n2, attr in graph.edges(data=True)}

	
	
	def get_graph_name(self, graph_id):
		"""
	/*!
	 * @brief Returns the graph name.
	 * @param[in] graph_id ID of an input graph that has been added to the environment.
	 * @return Name of the input graph.
	 */
		"""
		return self._ged_data._graph_names[graph_id]
	
	
	def get_graph_num_nodes(self, graph_id):
		"""
	/*!
	 * @brief Returns the number of nodes.
	 * @param[in] graph_id ID of an input graph that has been added to the environment.
	 * @return Number of nodes in the graph.
	 */
		"""
		return nx.number_of_nodes(self._ged_data.graph(graph_id))
	
	
	def get_original_node_ids(self, graph_id):
		"""
			Searchs and returns all th Ids of nodes on a graph, selected by its ID. 
	
			:param graph_id: The ID of the wanted graph
			:type graph_id: size_t
			:return: The list of IDs's nodes on the selected graph
			:rtype: list[string]
			
			.. seealso::get_graph_internal_id(), get_graph_num_nodes(), get_graph_num_edges(), get_graph_node_labels(), get_graph_edges(), get_graph_adjacence_matrix()
			.. note:: These functions allow to collect all the graph's informations.
		"""
		return [i for i in self._internal_to_original_node_ids[graph_id].values()]		
	
	
	def get_node_cost(self, node_label_1, node_label_2):
		return self._ged_data.node_cost(node_label_1, node_label_2)
	
	
	def get_node_rel_cost(self, node_label_1, node_label_2):
		"""
	/*!
	 * @brief Returns node relabeling cost.
	 * @param[in] node_label_1 First node label.
	 * @param[in] node_label_2 Second node label.
	 * @return Node relabeling cost for the given node labels.
	 */
		"""
		if isinstance(node_label_1, dict):
			node_label_1 = tuple(sorted(node_label_1.items(), key=lambda kv: kv[0]))
		if isinstance(node_label_2, dict):
			node_label_2 = tuple(sorted(node_label_2.items(), key=lambda kv: kv[0]))
		return self._ged_data._edit_cost.node_rel_cost_fun(node_label_1, node_label_2) # @todo: may need to use node_cost() instead (or change node_cost() and modify ged_method for pre-defined cost matrices.)
		
		
	def get_node_del_cost(self, node_label):
		"""
	/*!
	 * @brief Returns node deletion cost.
	 * @param[in] node_label Node label.
	 * @return Cost of deleting node with given label.
	 */
		"""
		if isinstance(node_label, dict):
			node_label = tuple(sorted(node_label.items(), key=lambda kv: kv[0]))
		return self._ged_data._edit_cost.node_del_cost_fun(node_label)
		
		
	def get_node_ins_cost(self, node_label):
		"""
	/*!
	 * @brief Returns node insertion cost.
	 * @param[in] node_label Node label.
	 * @return Cost of inserting node with given label.
	 */
		"""
		if isinstance(node_label, dict):
			node_label = tuple(sorted(node_label.items(), key=lambda kv: kv[0]))
		return self._ged_data._edit_cost.node_ins_cost_fun(node_label)
	
	
	def get_edge_cost(self, edge_label_1, edge_label_2):
		return self._ged_data.edge_cost(edge_label_1, edge_label_2)
		
		
	def get_edge_rel_cost(self, edge_label_1, edge_label_2):
		"""
	/*!
	 * @brief Returns edge relabeling cost.
	 * @param[in] edge_label_1 First edge label.
	 * @param[in] edge_label_2 Second edge label.
	 * @return Edge relabeling cost for the given edge labels.
	 */
		"""
		if isinstance(edge_label_1, dict):
			edge_label_1 = tuple(sorted(edge_label_1.items(), key=lambda kv: kv[0]))
		if isinstance(edge_label_2, dict):
			edge_label_2 = tuple(sorted(edge_label_2.items(), key=lambda kv: kv[0]))
		return self._ged_data._edit_cost.edge_rel_cost_fun(edge_label_1, edge_label_2)
		
		
	def get_edge_del_cost(self, edge_label):
		"""
	/*!
	 * @brief Returns edge deletion cost.
	 * @param[in] edge_label Edge label.
	 * @return Cost of deleting edge with given label.
	 */
		"""
		if isinstance(edge_label, dict):
			edge_label = tuple(sorted(edge_label.items(), key=lambda kv: kv[0]))
		return self._ged_data._edit_cost.edge_del_cost_fun(edge_label)
		
		
	def get_edge_ins_cost(self, edge_label):
		"""
	/*!
	 * @brief Returns edge insertion cost.
	 * @param[in] edge_label Edge label.
	 * @return Cost of inserting edge with given label.
	 */
		"""
		if isinstance(edge_label, dict):
			edge_label = tuple(sorted(edge_label.items(), key=lambda kv: kv[0]))
		return self._ged_data._edit_cost.edge_ins_cost_fun(edge_label)
		
		
	def get_all_graph_ids(self):
		return [i for i in range(0, self._ged_data._num_graphs_without_shuffled_copies)]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul  7 11:30:31 2020

@author: ljia
"""
import numpy as np
import cvxpy as cp
import time
from gklearn.utils import Timer


class CostsLearner(object):
	
	
	def __init__(self, parallel, verbose):
		### To set.
		self._parallel = parallel
		self._verbose = verbose
		# For update().
		self._time_limit_in_sec = 0
		self._max_itrs = 100
		self._max_itrs_without_update = 3
		self._epsilon_residual = 0.01
		self._epsilon_ec = 0.1
		### To compute.
		self._residual_list = []
		self._runtime_list = []
		self._cost_list = []
		self._nb_eo = None
		# For update().
		self._itrs = 0
		self._converged = False
		self._num_updates_ecs = 0
		self._ec_changed = None
		self._residual_changed = None
		self._itrs_without_update = 0
		### Both set and get.
		self._ged_options = None


	def fit(self, X, y):
		pass
	
	
	def preprocess(self):
		pass # @todo: remove the zero numbers of edit costs.
	
	
	def postprocess(self):
		for i in range(len(self._cost_list[-1])):
			if -1e-9 <= self._cost_list[-1][i] <= 1e-9:
				self._cost_list[-1][i] = 0
			if self._cost_list[-1][i] < 0:
				raise ValueError('The edit cost is negative.')
	
	
	def set_update_params(self, **kwargs):
		self._time_limit_in_sec = kwargs.get('time_limit_in_sec', self._time_limit_in_sec)
		self._max_itrs = kwargs.get('max_itrs', self._max_itrs)
		self._max_itrs_without_update = kwargs.get('max_itrs_without_update', self._max_itrs_without_update)
		self._epsilon_residual = kwargs.get('epsilon_residual', self._epsilon_residual)
		self._epsilon_ec = kwargs.get('epsilon_ec', self._epsilon_ec)
		

	def update(self, y, graphs, ged_options, **kwargs):
		# Set parameters.
		self._ged_options = ged_options
		if kwargs != {}:
			self.set_update_params(**kwargs)
		
		# The initial iteration.
		if self._verbose >= 2:
			print('\ninitial:')
		self.init_geds_and_nb_eo(y, graphs)

		self._converged = False
		self._itrs_without_update = 0
		self._itrs = 0
		self._num_updates_ecs = 0
		timer = Timer(self._time_limit_in_sec)
		# Run iterations from initial edit costs.
		while not self.termination_criterion_met(self._converged, timer, self._itrs, self._itrs_without_update):
			if self._verbose >= 2:
				print('\niteration', self._itrs + 1)
			time0 = time.time()
				
			# Fit GED space to the target space.
			self.preprocess()
			self.fit(self._nb_eo, y)
			self.postprocess()
			
			# Compute new GEDs and numbers of edit operations.
			self.update_geds_and_nb_eo(y, graphs, time0)
			
			# Check convergency.
			self.check_convergency()
			
			# Print current states.
			if self._verbose >= 2:
				self.print_current_states()
				
			self._itrs += 1
			
			
	def init_geds_and_nb_eo(self, y, graphs):
		pass
	
	
	def update_geds_and_nb_eo(self, y, graphs, time0):
		pass
			
			
	def compute_geds_and_nb_eo(self, graphs):
		pass
	
	
	def check_convergency(self):
		pass
	
	
	def print_current_states(self):
		pass


	def termination_criterion_met(self, converged, timer, itr, itrs_without_update):
		if timer.expired() or (itr >= self._max_itrs if self._max_itrs >= 0 else False):
# 			if self._state == AlgorithmState.TERMINATED:
# 				self._state = AlgorithmState.INITIALIZED
			return True
		return converged or (itrs_without_update > self._max_itrs_without_update if self._max_itrs_without_update >= 0 else False)
	
	
	def execute_cvx(self, prob):
		try:
			prob.solve(verbose=(self._verbose>=2))
		except MemoryError as error0:
			if self._verbose >= 2:
				print('\nUsing solver "OSQP" caused a memory error.')
				print('the original error message is\n', error0)
				print('solver status: ', prob.status)
				print('trying solver "CVXOPT" instead...\n')
			try:
				prob.solve(solver=cp.CVXOPT, verbose=(self._verbose>=2))
			except Exception as error1:
				if self._verbose >= 2:
					print('\nAn error occured when using solver "CVXOPT".')
					print('the original error message is\n', error1)
					print('solver status: ', prob.status)
					print('trying solver "MOSEK" instead. Notice this solver is commercial and a lisence is required.\n')
				prob.solve(solver=cp.MOSEK, verbose=(self._verbose>=2))
			else:
				if self._verbose >= 2:
					print('solver status: ', prob.status)					
		else:
			if self._verbose >= 2:
				print('solver status: ', prob.status)
		if self._verbose >= 2:				
			print()
			
			
	def get_results(self):
		results = {}
		results['residual_list'] = self._residual_list
		results['runtime_list'] = self._runtime_list
		results['cost_list'] = self._cost_list
		results['nb_eo'] = self._nb_eo
		results['itrs'] = self._itrs
		results['converged'] = self._converged
		results['num_updates_ecs'] = self._num_updates_ecs
		results['ec_changed'] = self._ec_changed
		results['residual_changed'] = self._residual_changed
		results['itrs_without_update'] = self._itrs_without_update
		return results